/**
 * @license Coati library
 * Lightweight, versatile JavaScript library to aid in cross-browser-compatible web sites.
 * (c) 2006 - 2012 by Yves Goergen
 * Last update: 2012-11-16
 * Website: http://unclassified.de
 * Licence: GNU General Public License (GPL) Version 3
 */

// Changes:
// 2012-11-16 - Zoomable image layout fixed, opacity fixes for IE

// This file contains almost the complete functionality of the following modules of the older
// dynamic-lib <http://beta.unclassified.de/code/html-css-js/dynamic-lib/>:
// * animation
// * color
// * debug
// * dom
// * dom.editor (improved, incl. auto-height)
// * dom.events
// * dom.keyevents
// * dom.selector
// * htmlform
// * infolabel (as popup, work in progress, no images)
// * runtime
// * webrequest
//
// The following more modules could be useful:
// * infolabel - rewrite with canvas, relying on excanvas for IE < 9
// * listbox - better rewrite as autocomplete-textbox
// * ui.effects - only the createBackgroundLayer part + a way to put some content onto it, like messages
// * textarea/input: live quotes replacement (auto-correct) (context-aware: don't replace anything inside {{{ ... }}})
//
// See the TODO file for dynamic-lib for more work to be done.

var coati = function()
{
	// **********************************************************************
	// ***************                                        ***************
	// ***************               MAIN CLASS               ***************
	// ***************                                        ***************
	// **********************************************************************

	var runtimeConfigStore = {};
	var showErrorMessages = true;
	var showErrorMessagesCounter = 0;
	var deferMemory = {};

	var self = {
		// Stores a configuration value in the local cache.
		//
		// This mechanism is used by other modules that can be configured.
		//
		setConfigValue: function(name, value)
		{
			runtimeConfigStore[name] = value;
		},

		// Retrieves a configuration value from the local cache.
		//
		getConfigValue: function(name, def)
		{
			var value = runtimeConfigStore[name];
			if (!isSet(value)) value = def;
			return value;
		},

		// Turns on/off displaying of error messages through the errorMessage function.
		//
		setErrorMessages: function(show)
		{
			showErrorMessages = !!show;
			showErrorMessagesCounter = 0;
		},

		isSet: function(o)
		{
			return typeof o != "undefined" && o != null;
		},

		isNumber: function(o)
		{
			return typeof o == "number";
		},

		isBoolean: function(o)
		{
			return typeof o == "boolean";
		},

		isString: function(o)
		{
			return typeof o == "string";
		},

		isArray: function(o)
		{
			return typeof o == "object" && o instanceof Array;
		},

		isObject: function(o)
		{
			return typeof o == "object";
		},

		isFunction: function(o)
		{
			return typeof o == "function";
		},

		isNumeric: function(o)
		{
			return isNumber(o) || isString(o) && o.search(/^-?[0-9]+(.[0-9]+)?$/) != -1;
		},

		isEven: function(x)
		{
			return isNumber(x) && x % 2 == 0;
		},

		isOdd: function(x)
		{
			return isNumber(x) && x % 2 == 1;
		},

		// This is not exactly JavaScript's definition of "truthy" values. It's more like PHP's.
		//
		isTrue: function(x)
		{
			if (!isSet(x)) return false;
			if (isString(x)) return coati_string.trim(x) != "" && coati_string.trim(x) != "0" && coati_string.trim(x).toLowerCase() != "false";
			if (isObject(x)) return x != null;
			if (isArray(x)) return x.length > 0;
			if (isNumber(x)) return x != 0;
			if (isBoolean(x)) return x;
			return false;
		},

		dump: function(v)
		{
			function dump_rec(v, level)
			{
				if (level > 5)
					return "{recursion}";

				var nl = "\n";
				for (var i = 0; i < level; i++)
					nl += "    ";
				var msg = "";
				if (!isSet(v))
					msg += "undefined";
				else if (isNumber(v))
					msg += v.toString();
				else if (isBoolean(v))
					msg += v.toString();
				else if (isString(v))
					msg += "\"" + v.toString().replace(/"/g, "\\\"") + "\"";
				else if (isArray(v))
				{
					msg += "[";
					for (var i = 0; i < v.length; i++)
					{
						msg += (i ? "," : "") + nl + "    ";
						if (msg.length > 500 && msg.indexOf("{toolong}") == -1)
						{
							msg += "{toolong}";
							break;
						}
						msg += i + ": ";
						try
						{
							msg += dump_rec(v[i], level + 1);
						}
						catch (ex)
						{
							msg += "{error}";
						}
					}
					msg += (i ? nl : " ") + "]";
				}
				else if (isObject(v))
				{
					msg += "{";
					var i = 0;
					for (var x in v) //if (v.hasOwnProperty(x))
					{
						msg += (i++ ? "," : "") + nl + "    ";
						if (msg.length > 500 && msg.indexOf("{toolong}") == -1)
						{
							msg += "{toolong}";
							break;
						}
						msg += x + ": ";
						try
						{
							msg += dump_rec(v[x], level + 1);
						}
						catch (ex)
						{
							msg += "{error}";
						}
					}
					msg += (i ? nl : " ") + "}";
				}
				else if (isFunction(v))
					msg += "function";
				else
					msg += "unknown";
				return msg;
			}

			return dump_rec(v, 0);
		},

		// Reports an error message.
		//
		errorMessage: function(msg, context)
		{
			if (showErrorMessages && msg != "")
			{
				// Only display up to 3 error messages simultaneously
				if (showErrorMessagesCounter >= 3) return;

				if (!isSet(context)) context = "";
				var AFn = "";
				for (var i = 0; i < 9 + (context.length * 0.85); i++) AFn += String.fromCharCode(0xAF);
				showErrorMessagesCounter++;
				alert((context ? "Context: " + context + "\n" + AFn + "\n" : "") + msg);
				showErrorMessagesCounter--;
			}
		},

		// Formats an exception object into a readable string
		//
		formatException: function(ex)
		{
			try
			{
				var nbsp = String.fromCharCode(160);
				var nbsp8 = nbsp + nbsp + nbsp + nbsp + nbsp + nbsp + nbsp + nbsp;
				var msg = "";
				msg += "Exception: " + ex.name + "\n";
				msg += "Message: " + ex.message + "\n";
				if (ex.fileName)
					msg += "Code: " + ex.fileName + ":" + ex.lineNumber + "\n";
				if (ex.stack)
					msg += "Stack:\n" + nbsp8 + coati_string.trim(ex.stack).replace(/\r/g, "").replace(/^((?:.*?\n){25}).*$/, "$1").replace(/(^|\n)(.{150}).*?(?=\n|$)/g, "$1$2").replace(/\n/g, "\n" + nbsp8) + "\n";
				return msg;
			}
			catch (ex2)
			{
				if (ex2.stack)
					return "Error formatting exception: " + ex2 + "\n\nInner exception: " + ex;
				else
					return "Error formatting exception:\n" + coati.dump(ex2) + "\nInner exception:\n" + coati.dump(ex);
			}
		},

		// Loads another JavaScript source file asynchronously (non-blocking).
		//
		includeScriptAsync: function(url)
		{
			var script = document.createElement("script");
			script.type = "text/javascript";
			script.src = url;
			document.body.appendChild(script);
		},

		// Loads another JavaScript source file synchronously (blocking).
		//
		includeScript: function(url)
		{
			// TODO: Use a synchronous HTTPRequest on the js file and eval() the returned code
			//       see webrequest.js
		},
//		// Evaluates dynamic JavaScript source code.
//		//
//		evalCode: function(code)
//		{
//			try
//			{
//				eval(code);
//			}
//			catch (ex)
//			{
//				coati.errorMessage("Exception caught while eval'ing code:\n" + coati.formatException(ex), "evalCode");
//			}
//		},

		// Defers a function call
		//
		// f = (function) Function to call deferred. Cannot be an anonymous inline function to use the clr parameter.
		// t = (int) Delay time in milliseconds
		// clr = (bool) First cancel a previously scheduled call of this very function
		//
		defer: function(f, t, clr)
		{
			var id = coati_array.findValue(deferMemory, f);
			if (clr && id)
			{
				clearTimeout(id);
				deferMemory[id] = undefined;
			}

			id = setTimeout(f, t);
			deferMemory[id] = f;
		},

		// Cuts off the filename from a URL to get its domain and path only
		//
		// Keeps the trailing slash (/) at the path value.
		//
		getURLDirectoryName: function(url)
		{
			return url.replace(/([^/]*)$/, "");
		},

		// Cuts off the domain and path from a URL to get its filename only
		//
		getURLFileName: function(url)
		{
			return url.replace(/^(.*\/)([^/]*)$/, "$2");
		},

		// Formats a decimal number
		//
		// num = (number) Number to format
		// dec = (int) Number of decimals to display. All if unset
		//
		formatNumber: function(num, dec)
		{
			if (isNaN(num)) return num;
			if (isSet(dec))
			{
				var factor = Math.pow(10, dec);
				num = Math.round(num * factor) / factor;
			}
			var str = num + "";
			if (dec > 0)
			{
				if (str.indexOf(".") == -1) str += ".";
				while (str.length - str.indexOf(".") - 1 < dec) str += "0";
			}
			return str.replace(/\./, coati.getConfigValue("i18n.decimal_separator", "."));
		},

		// Parses a numeric value in a string, with localised decimal separator.
		//
		parseNumber: function(s)
		{
			s = s.replace(coati.getConfigValue("i18n.decimal_separator", "."), ".");
			return +s;
		},


		// **********************************************************************
		// ***************                                        ***************
		// ***************           BROWSER DETECTION            ***************
		// ***************                                        ***************
		// **********************************************************************

		browser: function()
		{
			return {
				// Returns the Internet Explorer version, or 0 for other browsers.
				//
				// Source: http://msdn.microsoft.com/en-us/library/ms537509%28v=vs.85%29.aspx#4
				//
				ieVersion: function()
				{
					var style = document.documentElement.style;
					if (style.scrollbar3dLightColor != undefined)
					{
						if (style.opacity != undefined)
							return 9;
						else if (style.msBlockProgression != undefined)
							return 8;
						else if (style.msInterpolationMode != undefined)
							return 7;
						else if (style.textOverflow != undefined)
							return 6;
						else
							return 5.5;
					}
					return 0;
				},

				// Determines whether this is an Internet Explorer running.
				//
				ie: function()
				{
					return coati.browser.ieVersion() > 0;
				},

				// Possibly more browsers to detect:
				// http://dean.edwards.name/weblog/2007/03/sniff/ - comments removed by Google Closure Compiler
				// http://webreflection.blogspot.com/2009/01/32-bytes-to-know-if-your-browser-is-ie.html
				// http://www.thespanner.co.uk/2009/01/29/detecting-browsers-javascript-hacks/
				// http://stackoverflow.com/questions/5127970/#5197217

				// Determines the name of the CSS transition style property.
				//
				// Returns undefined if the browser doesn't support CSS transitions.
				//
				transition: function()
				{
					if (isSet(document.body.style.MozTransition))
					{
						return "MozTransition";
					}
					else if (isSet(document.body.style.WebkitTransition))
					{
						return "WebkitTransition";
					}
					else if (isSet(document.body.style.OTransition))
					{
						return "OTransition";
					}
					return undefined;
				},
				
				setCookie: function(cookieName, value, expireSeconds, path, domain, secure)
				{
					var expiryDate;
					if (expireSeconds)
					{
						expiryDate = new Date();
						expiryDate.setTime(expiryDate.getTime() + expireSeconds * 1000);
					}

					document.cookie = cookieName + '=' + encodeURIComponent(value) +
						(expireSeconds ? ';expires=' + expiryDate.toGMTString() : '') +
						';path=' + (path || '/') +
						(domain ? ';domain=' + domain : '') +
						(secure ? ';secure' : '');
				},
				
				deleteCookie: function(cookieName, path, domain, secure)
				{
					document.cookie = cookieName + '=;expires=0' +
						';path=' + (path || '/') +
						(domain ? ';domain=' + domain : '') +
						(secure ? ';secure' : '');
				},
				
				getCookie: function(cookieName)
				{
					var cookiePattern = new RegExp('(^|;)[ ]*' + cookieName + '=([^;]*)'),
					cookieMatch = cookiePattern.exec(document.cookie);
					return cookieMatch ? decodeURIComponent(cookieMatch[2]) : undefined;
				}
			};
		}(),


		// **********************************************************************
		// ***************                                        ***************
		// ***************             ARRAY HANDLING             ***************
		// ***************                                        ***************
		// **********************************************************************

		array: function()
		{
			return {
				// Gets the index of a value's first occurence in the array
				//
				// a = (array) Array to search in
				// o = (any) Value to search for
				// s = (int) Optional start index, if not 0
				//
				// NOTE: Firefox' built-in implementation:
				//       http://developer.mozilla.org/en/docs/Core_JavaScript_1.5_Reference:Objects:Array:indexOf
				//
				indexOf: function(a, o, s)
				{
					if (!isSet(a)) return -1;
					if (!isSet(s)) s = 0;
					if (s < 0) s += a.length;
					if (s < 0) s = 0;   // Limit start index to the first item
					for (; s < a.length; s++)
						if (a[s] === o) return s;
					return -1;
				},

				// Gets the index of a value's last occurence in the array
				//
				// a = (array) Array to search in
				// o = (any) Value to search for
				// s = (int) Optional start index, if not 0
				//
				// NOTE: Firefox' built-in implementation:
				//       http://developer.mozilla.org/en/docs/Core_JavaScript_1.5_Reference:Objects:Array:lastIndexOf
				//
				lastIndexOf: function(a, o, s)
				{
					if (!isSet(a)) return -1;
					var l = a.length;
					if (!isSet(s)) s = l - 1;
					if (s < 0) s += l;
					if (s > l - 1) s = l - 1;   // Limit start index to the last item
					for (; s >= 0; s--)
						if (a[s] === o) return s;
					return -1;
				},

				// Tests whether the array contains a given value
				//
				contains: function(a, o)
				{
					if (!isSet(a)) return false;
					for (var i = 0; i < a.length; i++)
						if (a[i] === o) return true;
					return false;
				},

				// Applies a function on each array element, without altering the array
				//
				// NOTE: Similar Firefox extension:
				//       http://developer.mozilla.org/en/docs/Core_JavaScript_1.5_Reference:Objects:Array:forEach
				//
				each: function(a, f)
				{
					if (!isSet(a)) return;
					for (var i = 0; i < a.length; i++)
						//if (i in a)   // [IN]
							f(a[i], i);
				},

				flatten: function(a)
				{
					if (!isSet(a)) return;
					var f = [];
					for (var i = 0; i < a.length; i++)
						//if (i in a)   // [IN]
							f = f.concat(a[i] instanceof Array ? coati_array.flatten(a[i]) : a[i]);
					return f;
				},

				// Applies a function on each array element and returns a new array with the function's return values
				//
				// NOTE: Firefox' built-in implementation:
				//       http://developer.mozilla.org/en/docs/Core_JavaScript_1.5_Reference:Objects:Array:map
				//
				map: function(a, f)
				{
					if (!isSet(a)) return;
					var m = new Array(a.length);
					for (var i = 0; i < a.length; i++)
						//if (i in a)   // [IN]
							m[i] = f(a[i], i);
					return m;
				},

				// Filters array elements by a custom function and returns the filtered array
				//
				// NOTE: Firefox' built-in implementation:
				//       http://developer.mozilla.org/en/docs/Core_JavaScript_1.5_Reference:Objects:Array:filter
				//
				filter: function(a, f)
				{
					if (!isSet(a)) return;
					var r = [];
					for (var i = 0; i < a.length; i++)
						if (/*i in a &&*/ f(a[i], i)) r.push(a[i]);   // [IN]
					return r;
				},

				// Removes an array element and returns the new array
				//
				remove: function(a, o)
				{
					return coati_array.filter(a, function(x) { return x != o });
				},

				// Removes an array element by its index and returns the new array
				//
				// Negative index values count from the end of the array: -1 is the last item.
				//
				// NOTE: JavaScript's Array.splice(index, count) can be used to remove array items.
				//       This only allows positive indices though.
				//
				removeAt: function(a, n)
				{
					if (!isSet(a)) return;
					if (n < 0) n += a.length;
					return coati_array.filter(a, function(x, i) { return n != i });
				},

				// Removes duplicate array elements and returns the new array
				//
				uniq: function(a)
				{
					if (!isSet(a)) return;
					var r = [];
					for (var i = 0; i < a.length; i++)
						if (/*i in a &&*/ !coati_array_contains(r, a[i]))   // [IN]
							r.push(a[i]);
					return r;
				},

				// Copies all array elements into a new array and returns the new array
				//
				clone: function(a)
				{
					if (!isSet(a)) return;
					var r = [];
					for (var i = 0; i < a.length; i++)
						//if (i in a)   // [IN]
							r[i] = a[i];
					return r;
				},

				// Determines whether none of this array's item functions return true for a given argument.
				//
				// TODO: Optionally take a callback function like in Firefox' Array.every
				//
				none: function(a, arg)
				{
					return !coati_array.any(a, arg);
				},

				// Determines whether any of this array's item functions returns true for a given argument.
				//
				// NOTE: Similar Firefox extension:
				//       http://developer.mozilla.org/en/docs/Core_JavaScript_1.5_Reference:Objects:Array:some
				//       (Requires a callback function)
				//
				// TODO: Optionally take a callback function like in Firefox' Array.some
				//
				any: function(a, arg)
				{
					if (!isSet(a)) return false;
					for (var i = 0; i < a.length; i++)
						if (/*i in a &&*/ a[i](arg)) return true;   // [IN]
					return false;
				},

				// Determines whether all of this array's item functions return true for a given argument.
				//
				// NOTE: Similar Firefox extension:
				//       http://developer.mozilla.org/en/docs/Core_JavaScript_1.5_Reference:Objects:Array:every
				//       (Requires a callback function)
				//
				// TODO: Optionally take a callback function like in Firefox' Array.every
				//
				all: function(a, arg)
				{
					if (!isSet(a)) return false;
					for (var i = 0; i < a.length; i++)
						if (/*i in a &&*/ !a[i](arg)) return false;   // [IN]
					return true;
				},

				// Tests whether the object contains a given key
				//
				// NOTE: Use the in operator instead
				//       http://developer.mozilla.org/en/docs/Core_JavaScript_1.5_Reference:Operators:Special_Operators:in_Operator
				//       Example: alert("someKey" in object)
				//
				//containsKey: function(a, o)
				//{
				//	for (var prop in a)
				//		if (prop === o) return true;
				//	return false;
				//},

				// Finds an object's property name that has the given value
				//
				findValue: function(a, o)
				{
					if (isSet(a))
						for (var prop in a)
							if (a[prop] === o) return prop;
				},

				// Tests whether the object contains a given value in one of its properties
				//
				containsValue: function(a, o)
				{
					return isSet(coati_array.findValue(a, o));
				}
			};
		}(),


		// **********************************************************************
		// ***************                                        ***************
		// ***************            STRING HANDLING             ***************
		// ***************                                        ***************
		// **********************************************************************

		string: function()
		{
			return {
				// Tests whether the string contains a given string
				//
				contains: function(s, str)
				{
					return s.indexOf(str) != -1;
				},

				// Trims all whitespace (or specified characters) from the beginning of the string
				//
				trimStart: function(s, c)
				{
					if (isSet(c))
					{
						for (var i = 0; i < s.length && coati_string.contains(c, s.charAt(i)); i++) { }
						return s.substr(i);
					}
					return s.replace(/^\s+/, "");
				},

				// Trims all whitespace (or specified characters) from the end of the string
				//
				trimEnd: function(s, c)
				{
					if (isSet(c))
					{
						for (var i = s.length - 1; i >= 0 && coati_string.contains(c, s.charAt(i)); i--) { }
						return s.substr(0, i + 1);   // 2nd argument = length
					}
					return s.replace(/\s+$/, "");
				},

				// Trims all whitespace (or specified characters) from the beginning and end of the string
				//
				trim: function(s, c)
				{
					return coati_string.trimEnd(coati_string.trimStart(s, c), c);
				},

				// Tests whether the string begins with a given string
				//
				startsWith: function(s, str)
				{
					if (!isString(str)) return false;
					return s.substr(0, str.length) == str;
				},

				// Tests whether the string ends with a given string
				//
				endsWith: function(s, str)
				{
					if (!isString(str)) return false;
					return s.substr(s.length - str.length, str.length) == str;
				},

				// Pads a string to a minimum length
				//
				// minLength = (int) minimum length
				// ch = (string) character/string to add. Default is " " (space)
				// atBegin = (bool) pad at the beginning (left padding). Default is at the end (right padding)
				//
				pad: function(s, minLength, ch, atBegin)
				{
					var r = s;
					if (!isSet(ch)) ch = " ";
					while (r.length < minLength)
						if (atBegin)
							r = ch + r;
						else
							r += ch;
					return r;
				},

				// Repeats a string n times
				//
				// Usage like: "hello".repeat(5)
				//
				repeat: function(s, n)
				{
					var r = "";
					while (n--) r += s;
					return r;
				},

				// Escapes special characters for XML (and HTML)
				//
				// s = (string) String to escape
				// quotes = (boolean) Also escape single and double quotes
				//
				escapeXml: function(s, quotes)
				{
					s = s.replace(/&/g, "&amp;").replace(/</g, "&lt;").replace(/>/g, "&gt;");
					if (quotes)
						s = s.replace(/"/g, "&quot;").replace(/'/g, "&#39;");
					if (0) s.replace(/"/g, "");   // NOTE: Fixes syntax highlighting for stupid text editors that don't know /.../.
					return s;
				},

				// Escapes special characters for regular expressions body
				//
				// This cannot be used inside character classes (which is the square brackets)! Use escapeRegexCC for that.
				//
				escapeRegex: function(s)
				{
					return s.replace(/([$^*+.?\\()/|\[\]{}])/g, "\\$1");
				},

				// Escapes special characters for regular expressions character classes
				//
				escapeRegexCC: function(s)
				{
					return s.replace(/([-\\^\[\]])/g, "\\$1");
				},

				// Escapes special characters for URIs
				//
				// This is different from encodeURIComponent() in that it replaces " " by "+" and does not touch commas (,).
				//
				encodeUri: function(s)
				{
					return encodeURIComponent(s).replace(/%20/g, "+").replace(/%2C/g, ",");
				},

				// Converts the string from hyphen-names to camelCase
				//
				// sep = (string) Optional separator character. Defaults to "-"
				//
				toCamelCase: function(s, sep)
				{
					sep = sep || "-";
					var parts = s.split(sep);
					var r = parts[0];
					for (var i = 1; i < parts.length; i++)
						r += parts[i].charAt(0).toUpperCase() + parts[i].substr(1);
					return r;
				}
			};
		}(),


		// **********************************************************************
		// ***************                                        ***************
		// ***************              DOM HANDLING              ***************
		// ***************                                        ***************
		// **********************************************************************

		dom: function()
		{
			var scrollTimeout = undefined;
			var scrollOffset = 0;
			var prevLocationHash;
			var historyChangedHandlers = [];
			var domLoaded = false;
			var pageLoaded = false;

			// Finds the event object
			//
			function findEvent(e)
			{
				e = e || event;   // IE
				if (!e.target && e.srcElement)   // IE
					e.target = e.srcElement;
				if (isSet(e.button))
				{
					// Every browser uses different values for the mouse buttons. Correct them here.
					// DOM says: 0 = left, 1 = middle, 2 = right (multiple buttons not supported)
					// Opera 7 and older and Konqueror are not specifically handled here.
					// See http://de.selfhtml.org/javascript/objekte/event.htm#button
					if (coati.browser.ie())
					{
						if (e.button & 1)
							e.mouseButton = 0;
						else if (e.button & 2)
							e.mouseButton = 2;
						else if (e.button & 4)
							e.mouseButton = 1;
					}
					else
						e.mouseButton = e.button;
				}
				return e;
			}

			// Finds the key code from an event object
			//
			function findKeyCode(e)
			{
				if (e.data) return e.data;   // W3C
				if (e.which) return e.which;   // Netscape
				return e.keyCode;   // IE, Firefox, etc
			}

			var self = {
				// ====================  ELEMENT GEOMETRY  ====================

				// Determines upper border position of the object
				//
				top: function(o)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return 0;
					var top = o.offsetTop;
					while (o = o.offsetParent) top += o.offsetTop;
					return top;
				},

				// Determines left border position of the object
				//
				left: function(o)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return 0;
					var left = o.offsetLeft;
					while (o = o.offsetParent) left += o.offsetLeft;
					// TODO: Test the following line.
					// Testcase: <html style="border-left: solid 10px #B0FF90;">
					//left -= parseInt(coati_dom.getEffectiveStyle(document.documentElement, "border-left-width"));
					return left;
				},

				// Determines bottom border position of the object
				//
				bottom: function(o)
				{
					o = coati_dom_select(o);
					return coati_dom.top(o) + coati_dom.height(o);
				},

				// Determines right border position of the object
				//
				right: function(o)
				{
					o = coati_dom_select(o);
					return coati_dom.left(o) + coati_dom.width(o);
				},

				// Determines width of the object
				//
				width: function(o)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return 0;
					return o.offsetWidth;
				},

				// Determines height of the object
				//
				height: function(o)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return 0;
					return o.offsetHeight;
				},

				// Vertically resizes an element
				//
				// cy = (int) Height offset
				// negative = (bool) Rezize to the top instead of the bottom. This is achieved by scrolling the window accordingly.
				//
				resizeVert: function(o, cy, negative)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return;
					var newCy = o.clientHeight + cy;
					var minCy = 0;
					var maxCy = 1000;
					if (o.tagName.toLowerCase() == "textarea") minCy = 60;
					if (newCy < minCy) newCy = minCy;
					if (newCy > maxCy) newCy = maxCy;
					var diffY = newCy - o.clientHeight;
					o.style.height = newCy + "px";

					// scroll page in the opposite offset direction to make the box resize at the top
					if (negative == true)
						window.scrollBy(0, diffY);
				},

				// Moves an object to a position
				//
				// x = (number) Horizontal position
				// y = (number) Vertical position
				//
				moveTo: function(o, x, y)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return;
					o.style.left = x + "px";
					o.style.top = y + "px";
				},

				// Moves an object by a number of pixels in X and Y direction
				//
				// dx = (number) Horizontal offset
				// dy = (number) Vertical offset
				//
				moveBy: function(o, dx, dy)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return;
					o.style.left = (parseInt(o.style.left) + dx) + "px";
					o.style.top = (parseInt(o.style.top) + dy) + "px";
				},

				// Resizes an object to a size
				//
				// w = (number) Horizontal size
				// h = (number) Vertical size
				//
				resizeTo: function(o, w, h)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return;
					o.style.width = w + "px";
					o.style.height = h + "px";
				},

				// Resizes an object by a number of pixels in X and Y direction
				//
				// dx = (number) Horizontal offset
				// dy = (number) Vertical offset
				//
				resizeBy: function(o, dx, dy)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return;
					o.style.width = (parseInt(o.style.width) + dx) + "px";
					o.style.height = (parseInt(o.style.height) + dy) + "px";
				},

				// Clears any width and height style set to let the object auto-size again
				//
				unsetSize: function(o)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return;
					o.style.width = "";
					o.style.height = "";
				},

				// Gets the inner client width of the browser window.
				//
				windowClientWidth: function()
				{
					if (document.documentElement.clientWidth > 0) return document.documentElement.clientWidth;
					if (isSet(document.body.clientWidth)) return document.body.clientWidth;
					if (isSet(window.innerWidth)) return window.innerWidth;
					return undefined;
				},

				// Gets the inner client height of the browser window.
				//
				windowClientHeight: function()
				{
					if (document.documentElement.clientHeight > 0) return document.documentElement.clientHeight;
					if (isSet(document.body.clientHeight)) return document.body.clientHeight;
					if (isSet(window.innerHeight)) return window.innerHeight;
					return undefined;
				},


				// ====================  ELEMENT STATE  ====================

				// Toggles visibility of an object
				//
				// state = (bool) true/false: Force to show/hide it [optional]
				//
				// Returns the new enabled state as -1 or +1
				//
				toggleVisible: function(o, state)
				{
					var newstate;
					o = coati_dom_select(o);
					if (!isSet(o)) return -1;
					if (!isSet(state))
						newstate = (o.style.display == "") ? "none" : "";
					else
						newstate = state ? "" : "none";
					o.style.display = newstate;
					return newstate == "" ? 1 : -1;
				},

				show: function(o)
				{
					return coati_dom.toggleVisible(o, true);
				},

				hide: function(o)
				{
					return coati_dom.toggleVisible(o, false);
				},

				visible: function(o)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return false;
					return o.style.display != "none";
				},

				// Toggles enabled state of an object
				//
				// state = (bool) true/false: Force to enable/disable it [optional]
				//
				// Returns the new enabled state as boolean
				//
				toggleEnabled: function(o, state)
				{
					var newstate;
					o = coati_dom_select(o);
					if (!isSet(o)) return false;
					if (!isSet(state))
						newstate = !o.disabled;
					else
						newstate = !state;
					o.disabled = newstate;
					return !newstate;
				},

				enable: function(o)
				{
					return coati_dom.toggleEnabled(o, true);
				},

				disable: function(o)
				{
					return coati_dom.toggleEnabled(o, false);
				},

				enabled: function(o)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return false;
					return !o.disabled;
				},

				// Disables an object in 50ms
				//
				disableAsync: function(o)
				{
					o = coati_dom_select(o);
					window.setTimeout(function() { if (o) { o.blur(); o.disabled = true; } }, 50);
				},

				// Toggles selection state of a button element, using the style class name "selected"
				//
				// Returns the new selected state as boolean
				//
				toggleSelected: function(o)
				{
					o = coati_dom_select(o);
					if (coati_dom.hasClass(o, "selected"))
					{
						coati_dom.removeClass(o, "selected");
						return false;
					}
					coati_dom.addClass(o, "selected");
					return true;
				},


				// ====================  WINDOW SCROLLING  ====================

				// Gets or sets the current horizontal scroll position
				//
				// Some references:
				//   http://www.evolt.org/article/document_body_doctype_switching_and_more/17/30655/ - Section about "Scrolling offset"
				//   http://www.howtocreate.co.uk/tutorials/javascript/browserwindow - Near the page bottom
				//
				scrollX: function(v)
				{
					if (isSet(window.pageXOffset))
						if (isSet(v))
						{
							window.scrollTo(v, coati_dom.scrollY());
							return v;
						}
						else
							return window.pageXOffset;
					// Else is IE, behaving differently depending on compatibility mode
					if (document.compatMode == "CSS1Compat")
						if (isSet(v))
							return document.documentElement.scrollLeft = v;
						else
							return document.documentElement.scrollLeft;
					else
						if (isSet(v))
							return document.body.scrollLeft = v;
						else
							return document.body.scrollLeft;
				},

				// Gets the current vertical scroll position
				//
				scrollY: function(v)
				{
					if (isSet(window.pageYOffset))
						if (isSet(v))
						{
							window.scrollTo(coati_dom.scrollX(), v);
							return v;
						}
						else
							return window.pageYOffset;
					// Else is IE, behaving differently depending on compatibility mode
					if (document.compatMode == "CSS1Compat")
						if (isSet(v))
							return document.documentElement.scrollTop = v;
						else
							return document.documentElement.scrollTop;
					else
						if (isSet(v))
							return document.body.scrollTop = v;
						else
							return document.body.scrollTop;
				},

				// Scrolls the window down by the height of an item
				//
				scrollDownBy: function(o)
				{
					o = coati_dom_select(o);
					window.scrollBy(0, coati_dom.height(o));
				},

				// Scrolls the window up by the height of an item
				//
				scrollUpBy: function(o)
				{
					o = coati_dom_select(o);
					window.scrollBy(0, -coati_dom.height(o));
				},

				// Sets the anchor scrolling vertical offset
				//
				setAnchorScrollOffset: function(y)
				{
					scrollOffset = y;
				},

				// Scrolls the window to the URL anchor
				//
				// The anchor must be defined on the page like <a id="myAnchor" />
				//
				scrollToAnchor: function()
				{
					// disable a running timeout if we have loaded faster than that
					try
					{
						clearTimeout(scrollTimeout);
					}
					catch (ex)
					{
					}

					var url = window.location.href;
					var id = "";
					var id_pos = url.lastIndexOf("#");
					if (id_pos > 0)
						id = url.substring(id_pos + 1, url.length);
					else
						// this is the place to match a custom URL...
					if (id != "")
					{
						var o = coati_dom_select(id);   // fetch the anchor element by its id attribute
						if (!isSet(o)) return;
						var top = coati_dom.top(o) - 7 + scrollOffset;
						window.scrollTo(0, top);
					}
				},

				// Scrolls the window to the position of an object
				//
				scrollToObject: function(o)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return;
					var top = coati_dom.top(o) - 7 + scrollOffset;
					window.scrollTo(0, top);
				},

				// Automatically scrolls the window to the URL anchor
				//
				// See scrollToAnchor function for details.
				// Automatically handles the document "load" event and tries to scroll
				// after a second already in case the page has not finished loading by then.
				//
				// Requires DOMEvents.
				//
				autoScrollToAnchor: function()
				{
					coati_dom.addEvent(window, "load", coati_dom.scrollToAnchor);
					scrollTimeout = setTimeout(coati_dom.scrollToAnchor, 1000);
				},

				// Toggles visibility of an object and scrolls the page up and down by its height
				//
				toggleVisibleScroll: function(o)
				{
					var newstate;
					o = coati_dom_select(o);
					if (!isSet(o)) return;

					if (o.style.display == "")
					{
						window.scrollBy(0, -coati_dom.height(o));
						newstate = "none";
						o.style.display = newstate;
					}
					else
					{
						newstate = "";
						o.style.display = newstate;
						window.scrollBy(0, coati_dom.height(o));
					}
					return (newstate == "" ? 1 : -1);
				},


				// ====================  ELEMENT CONTENT  ====================

				// Sets an image's source attribute depending on a condition
				//
				// This condition can be the return value of toggleVisId, e.g.
				//
				// cond = (bool) Condition
				// trueimg = (string) True case image source
				// falseimg = (string) False case image source
				//
				setImageCond: function(o, cond, trueimg, falseimg)
				{
					coati_dom_select(o).src = cond ? trueimg : falseimg;
				},


				// ====================  STYLESHEETS  ====================

				// Determines whether the object has one or multiple class names set
				//
				hasClass: function(o, n)
				{
					o = coati_dom_select(o);
					if (!isSet(o) || !isSet(o.className)) return false;
					var cn = o.className.split(" ");
					var req = n.split(" ");
					req = coati_array.filter(req, function(x) { return !coati_array_contains(cn, x); });
					return req.length == 0;
				},

				// Gets all class names of an element
				//
				getClasses: function(o)
				{
					o = coati_dom_select(o);
					if (!isSet(o) || !isSet(o.className)) return "";
					return o.className.split(" ");
				},

				// Adds one or multiple class names
				//
				addClass: function(o, n)
				{
					o = coati_dom_select(o);
					if (!isSet(o) || !isSet(o.className)) return;
					var cn = o.className.split(" ");
					var req = n.split(" ");
					coati_array_each(req, function(x) { if (x.length > 0 && !coati_array_contains(cn, x)) cn.push(x); });
					o.className = cn.join(" ");
				},

				// Removes one or multiple class names
				//
				removeClass: function(o, n)
				{
					o = coati_dom_select(o);
					if (!isSet(o) || !isSet(o.className)) return;
					var cn = o.className.split(" ");
					var req = n.split(" ");
					cn = coati_array.filter(cn, function(x) { return !coati_array_contains(req, x); });
					o.className = cn.join(" ");
				},

				// Toggle one or multiple CSS class names
				//
				// Each class name is toggled separately, i.e. independently of the others.
				//
				toggleClass: function(o, n, state)
				{
					o = coati_dom_select(o);
					if (!isSet(o) || !isSet(o.className)) return;
					var req = n.split(" ");
					coati_array_each(req, function(n)
					{
						if (state > 0 || !isSet(state) && !coati_dom.hasClass(o, n))
							coati_dom.addClass(o, n);
						else
							coati_dom.removeClass(o, n);
					});
				},

				// Adds a document-wide style definition in a new <style> tag in the <head> area
				//
				addPageStyle: function(css)
				{
					var style = document.createElement("style");
					style.setAttribute("type", "text/css");
					try
					{
						//style.innerHTML = css;
						style.appendChild(document.createTextNode(css));   // TODO: either way, this makes IE cry silently
					}
					catch (ex)
					{
						coati.errorMessage(coati.formatException(ex), "coati_dom.addPageStyle");
					}
					document.getElementsByTagName("head")[0].appendChild(style);
				},

				// Returns the effective (computed) style attribute for an element
				//
				// Specify the attribute name in hyphened notation, e.g. "background-color". It will be
				// converted to camelCase notation where required.
				//
				getEffectiveStyle: function(o, n)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return undefined;
					//if (document.defaultView && document.defaultView.getComputedStyle)   // DOM
					//	return document.defaultView.getComputedStyle(o, "").getPropertyValue(n);
					if (window.getComputedStyle)   // DOM
						return window.getComputedStyle(o, null).getPropertyValue(n);
						// null parameter required for Firefox up to 3.6, see https://developer.mozilla.org/en/DOM:window.getComputedStyle
					if (o.currentStyle)   // IE
						return o.currentStyle[coati_string.toCamelCase(n)];
					errorMessage("Browser does not support retrieval of effective style", "coati.dom.getEffectiveStyle");
				},

				// Gets the current opacity of an element (0.0 to 1.0)
				//
				getOpacity: function(o)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return undefined;
					if (o.filters && o.filters.alpha && isNumber(o.filters.alpha.opacity))   // IE
						return o.filters.alpha.opacity / 100;
					if (isSet(o.style.opacity) && o.style.opacity != "")   // others
						return o.style.opacity;
					return 1;
				},

				// Sets an element's opacity
				//
				// This feature isn't available in Opera 8.5 and before.
				// In Internet Explorer, this may remove other filters from the element (untested)
				//
				// opacity = (number) 0.0 (transparent) to 1.0 (opaque)
				//
				setOpacity: function(o, opacity)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return;
					if (isSet(o.style.opacity))   // CSS3
						o.style.opacity = opacity;
					else if (isSet(o.style.filter))   // IE
						o.style.filter = (opacity === 1) ? "" : "Alpha(opacity=" + (opacity * 100) + ")";
				},


				// ====================  SIMPLE ELEMENT SELECTION  ====================

				// Ascends the DOM tree starting from obj to find a node of name
				//
				// n = (string) Tag name to search
				//
				findParentByName: function(o, n)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return undefined;
					while (o.nodeName.toLowerCase() != n)
					{
						if (!o.parentNode) return undefined;
						o = o.parentNode;
					}
					return o;
				},

				// Ascends the DOM tree starting from obj to find a node with a class
				//
				// n = (string) Class name to search
				//
				findParentByClass: function(o, n)
				{
					o = coati_dom_select(o);
					if (!isSet(o)) return undefined;
					while (!coati_dom.hasClass(o, n))
					{
						if (!o.parentNode) return undefined;
						o = o.parentNode;
					}
					return o;
				},


				// ====================  ADVANCED ELEMENT SELECTION  ====================

				// Selects an element that matches the specified CSS selector query.
				//
				// If the first argument is not a string, it is returned immediately. This is the pass-through
				// for HTMLElement objects that don't need to be selected but already are a node object.
				//
				// q = (string) CSS selector query
				// o = (object/id) Where to start the search, array accepted
				//
				// Returns the matching element if one was found, otherwise undefined.
				//
				select: function(q, o)
				{
					// HTMLElement pass-through
					if (!isString(q)) return q;

					var a = coati_dom.selectAll(q, o);
					if (isSet(a) && a.length >= 1) return a[0];
				},

				// Selects all elements that match the specified CSS selector query.
				//
				// "#id"-only-style queries with no start node use document.getElementById. This makes it easy to
				// efficiently use this function as a replacement for select(). The single node is put in an array
				// anyway, so that this function always returns an array. Use the select() function above to only
				// get one node (i.e. the first of multiple).
				//
				// If the first argument is not a string, it is returned immediately. This is the pass-through
				// for HTMLElement objects that don't need to be selected but already are a node object.
				//
				// q = (string) CSS selector query
				// o = (object/id) Where to start the search, array accepted
				//
				// Returns an array of all matching elements.
				//
				// TODO: Pseudo-classes probably worth implementing:
				//       :nth-child()
				//       :nth-last-child()
				//       :nth-of-type()
				//       :nth-last-of-type()
				//       The :not() pseudo-class is a bit more complicated.
				//
				// CSS selectors quick reference of all implemented features:
				//
				// CSS2.1: http://www.w3.org/TR/CSS21/selector.html
				// CSS3: http://www.w3.org/TR/css3-selectors/
				//
				// *                 Matches any element (the * character is optional)
				// E                 an E element
				// E F               an F element that is a descendant of an E element
				// E > F             an F element that is a child of an E element
				// E + F             an F element immediately preceded by a sibling E element
				// E ~ F             an F element preceded by a sibling E element
				// E:first-child     an E element which is the first child of its parent
				// E:last-child      an E element which is the last child of its parent
				// E:first-of-type   an E element which is the first sibling of its type
				// E:last-of-type    an E element which is the last sibling of its type
				// E:only-child      an E element which is the only child of its parent
				// E:only-of-type    an E element which is the only sibling of its type
				// E:empty           an E element that has no children (including text nodes)
				// E:enabled         a user interface element E which is enabled
				// E:disabled        a user interface element E which is disabled
				// E:checked         a user interface element E which is checked (for instance a radio-button or checkbox)
				// E:contains("foo") an E element containing the substring "foo" in its textual contents
				// E[foo]            an E element with the "foo" attribute set
				// E[foo="bar"]      an E element whose "foo" attribute value is exactly equal to "bar"
				// E[foo~="bar"]     an E element whose "foo" attribute value is a list of space-separated values, one of which is exactly equal to "bar"
				// E[lang|="en"]     an E element whose "lang" attribute has a hyphen-separated list of values beginning (from the left) with "en"
				// E[foo^="bar"]     an E element whose "foo" attribute value begins exactly with the string "bar"
				// E[foo$="bar"]     an E element whose "foo" attribute value ends exactly with the string "bar"
				// E[foo*="bar"]     an E element whose "foo" attribute value contains the substring "bar"
				// E.cls             an E element with class "cls" assigned; the same as E[class~="cls"]
				// E#myid            an E element with ID equal to "myid"; the same as E[id="myid"]
				// E, F              an E or F element
				//
				selectAll: function(q, o)
				{
					// HTMLElement pass-through
					if (!isString(q))
						return q;

					if (isString(o))
						o = document.getElementById(o);
					if (!isSet(o))
						o = document;
					q = coati_string.trim(q);

					// "#id"-only-style optimisation
					var m;
					if (!isSet(o) && (m = q.match(/^#([-_0-9A-Za-z]+)$/)))
						return [document.getElementById(m[1])];

					// ".class"-only-style optimisation, for newer browsers
					if (o.getElementsByClassName && (m = q.match(/^\.([-_0-9A-Za-z]+)$/)))
						return o.getElementsByClassName(m[1]);

					if (isArray(o))
					{
						var nodes = [];
						for (var oi = 0; oi < o.length; oi++)
							nodes = nodes.concat(arguments.callee(q, o[oi]));
						return coati_array.uniq(nodes);
					}

					// First, parse the entire CSS selector query string into groups and elements with their attributes.
					var groups = [];
					var elements = [];
					while (q.length > 0)
					{
						// Read grouper
						if (elements.length > 0)
						{
							m = q.match(/^\s*,\s*/);
							if (isSet(m))
							{
								// Start a new group
								groups.push(elements);
								elements = [];
								q = q.substr(m[0].length);
							}
						}

						// Read combiner
						var comb = "";
						if (elements.length > 0)
						{
							m = q.match(/^(\s*[>+~]\s*|\s+)/);
							if (isSet(m))
							{
								comb = coati_string.trim(m[1]);
								q = q.substr(m[0].length);
							}
							else
								return false;
						}

						// Read element name
						var element = "";
						m = q.match(/^([-_0-9A-Za-z]+|\*)/);
						if (isSet(m))
						{
							element = m[1].toLowerCase();
							q = q.substr(m[0].length);
						}

						// Read specifiers
						var specs = [];
						do
						{
							// ID specification
							m = q.match(/^#([-_0-9A-Za-z]+)/);
							if (isSet(m))
								q = "[id=" + m[1] + "]" + q.substr(m[0].length);   // Continue with other handler

							// Class specification
							m = q.match(/^\.([-_0-9A-Za-z]+)/);
							if (isSet(m))
								q = "[class~=" + m[1] + "]" + q.substr(m[0].length);   // Continue with other handler

							// Attribute specifications
							m = q.match(/^\[([-_0-9A-Za-z]+)(?:(=|~=|\|=|\^=|\$=|\*=)(?:([^'"][^\]]*)|'((?:[^']|\\')*)'|"((?:[^"]|\\")*)"))?\]/);
							if (isSet(m))
							{
								// Need a new closure for each iteration to remember the correct attrValue, regex etc.
								(function() {
									var attrName = m[1];
									var op = m[2];
									var val = "";
									if (isSet(m[3]) && m[3] != "") val = m[3];   // No quotes
									else if (isSet(m[4]) && m[4] != "") val = m[4].replace(/\\'/g, "'");   // Single quotes - http://www.w3.org/TR/CSS21/syndata.html#strings
									else if (isSet(m[5]) && m[5] != "") val = m[5].replace(/\\"/g, '"');   // Double quotes
									q = q.substr(m[0].length);

									var regex;
									if (op == "=") regex = new RegExp("^" + coati_string.escapeRegex(val) + "$");   // exact match
									else if (op == "~=") regex = new RegExp("(^|\\s)" + coati_string.escapeRegex(val) + "(\\s|$)");   // space-separated list
									else if (op == "|=") regex = new RegExp("^" + coati_string.escapeRegex(val) + "(-|$)");   // beginning of hyphen-separated list
									else if (op == "^=") regex = new RegExp("^" + coati_string.escapeRegex(val));   // beginning of string
									else if (op == "$=") regex = new RegExp(coati_string.escapeRegex(val) + "$");   // end of string
									else if (op == "*=") regex = new RegExp(coati_string.escapeRegex(val));   // anywhere in string
									specs.push(function(node)
									{
				//						// NOTE: This code variant does not work for "id" and "style" attributes in IE; they're always set!
				//						if (!node.getAttribute) return false;
				//						// The "class" attribute needs special treatment, obviously
				//						if (attrName == "class")
				//							return isSet(regex) && regex.test(node.className) || !isSet(regex) && isSet(node.className);
				//						return isSet(regex) && regex.test(node.getAttribute(attrName)) ||
				//							!isSet(regex) && isSet(node.getAttribute(attrName));

										return node.attributes &&
											node.attributes[attrName] &&
											(isSet(regex) && regex.test(node.attributes[attrName].value) ||
											!isSet(regex) && node.attributes[attrName].specified);
									});
								})();
								continue;
							}

							// Match pseudo classes
							m = q.match(/^:first-child/);
							if (isSet(m))
							{
								q = q.substr(m[0].length);
								specs.push(function(node)
								{
									var pcn = node.parentNode.childNodes;
									for (var i = 0; i < pcn.length; i++)
										if (pcn[i].nodeType == 1)
											return pcn[i] == node;
								});
								continue;
							}

							m = q.match(/^:last-child/);
							if (isSet(m))
							{
								q = q.substr(m[0].length);
								specs.push(function(node)
								{
									var pcn = node.parentNode.childNodes;
									for (var i = pcn.length - 1; i >= 0; i--)
										if (pcn[i].nodeType == 1)
											return pcn[i] == node;
								});
								continue;
							}

							m = q.match(/^:first-of-type/);
							if (isSet(m))
							{
								q = q.substr(m[0].length);
								specs.push(function(node)
								{
									var pcn = node.parentNode.childNodes;
									for (var i = 0; i < pcn.length; i++)
										if (pcn[i].nodeType == 1 &&
											pcn[i].tagName.toLowerCase() == node.tagName.toLowerCase())
											return pcn[i] == node;
								});
								continue;
							}

							m = q.match(/^:last-of-type/);
							if (isSet(m))
							{
								q = q.substr(m[0].length);
								specs.push(function(node)
								{
									var pcn = node.parentNode.childNodes;
									for (var i = pcn.length - 1; i >= 0; i--)
										if (pcn[i].nodeType == 1 &&
											pcn[i].tagName.toLowerCase() == node.tagName.toLowerCase())
											return pcn[i] == node;
								});
								continue;
							}

							m = q.match(/^:only-child/);
							if (isSet(m))
							{
								q = q.substr(m[0].length);
								specs.push(function(node)
								{
									var pcn = node.parentNode.childNodes;
									var c = 0;
									for (var i = 0; i < pcn.length; i++)
										if (pcn[i].nodeType == 1)
											c++;
									return c == 1;
								});
								continue;
							}

							m = q.match(/^:only-of-type/);
							if (isSet(m))
							{
								q = q.substr(m[0].length);
								specs.push(function(node)
								{
									var pcn = node.parentNode.childNodes;
									var c = 0;
									for (var i = 0; i < pcn.length; i++)
										if (pcn[i].nodeType == 1 &&
											pcn[i].tagName.toLowerCase() == node.tagName.toLowerCase())
											c++;
									return c == 1;
								});
								continue;
							}

							m = q.match(/^:empty/);
							if (isSet(m))
							{
								q = q.substr(m[0].length);
								specs.push(function(node)
								{
									return node.childNodes.length == 0;
								});
								continue;
							}

							m = q.match(/^:enabled/);
							if (isSet(m))
							{
								q = q.substr(m[0].length);
								specs.push(function(node)
								{
									return !node.disabled;
								});
								continue;
							}

							m = q.match(/^:disabled/);
							if (isSet(m))
							{
								q = q.substr(m[0].length);
								specs.push(function(node)
								{
									return node.disabled;
								});
								continue;
							}

							m = q.match(/^:checked/);
							if (isSet(m))
							{
								q = q.substr(m[0].length);
								specs.push(function(node)
								{
									return node.checked;
								});
								continue;
							}

							m = q.match(/^:contains\((?:([^'"][^\]]*)|'((?:[^']|\\')*)'|"((?:[^"]|\\")*)")\)/);
							if (isSet(m))
							{
								// Need a new closure for each iteration to remember the correct text value
								(function() {
									var val = "";
									if (isSet(m[1]) && m[1] != "") val = m[1];   // No quotes
									else if (isSet(m[2]) && m[2] != "") val = m[2].replace(/\\'/g, "'");   // Single quotes
									else if (isSet(m[3]) && m[3] != "") val = m[3].replace(/\\"/g, '"');   // Double quotes
									q = q.substr(m[0].length);

									specs.push(function(node)
									{
										return (isSet(node.innerText) ? node.innerText : node.textContent).indexOf(val) != -1;
									});
								})();
								continue;
							}

						}
						while (isSet(m));

						if (element == "" && specs.length == 0)
							return false;   // No element and no specifiers... something is wrong!
						if (element == "")
							element = "*";

						elements.push({"comb": comb, "element": element, "specs": specs});
					} // while q.length

					if (elements.length > 0)
						groups.push(elements);

					var checkNode = function(node, elements)
					{
						if (elements.length === 1) return true;   // Accept if there are no further conditions (the last one would be removed)
						pe = elements[elements.length - 1];
						elements = elements.slice(0, -1);
						e = elements[elements.length - 1];

						if (pe.comb === "")   // Find any parent node
						{
							while (node.parentNode &&
								!(
									(e.element === "*" || node.parentNode.nodeName.toLowerCase() === e.element) &&
									coati_array.all(e.specs, node.parentNode) &&
									checkNode(node.parentNode, elements)
								))
								node = node.parentNode;
							return !!node.parentNode;
						}
						if (pe.comb === ">")   // Find direct parent node
						{
							return (node.parentNode &&
								(e.element === "*" || node.parentNode.nodeName.toLowerCase() === e.element) &&
								coati_array.all(e.specs, node.parentNode) &&
								checkNode(node.parentNode, elements));
						}
						if (pe.comb === "~")   // Find any previous sibling
						{
							while (node.previousSibling &&
								!(
									node.previousSibling.nodeType === 1 &&   // Must be an element node
									(e.element === "*" || node.previousSibling.nodeName.toLowerCase() === e.element) &&
									coati_array.all(e.specs, node.previousSibling) &&
									checkNode(node.previousSibling, elements)
								))
								node = node.previousSibling;
							return !!node.previousSibling;
						}
						if (pe.comb === "+")   // Find direct previous sibling
						{
							while (node.previousSibling && node.previousSibling.nodeType !== 1)
								node = node.previousSibling;   // Skip to the next element node
							return (node.previousSibling &&
								node.previousSibling.nodeType === 1 &&   // Must be an element node
								(e.element === "*" || node.previousSibling.nodeName.toLowerCase() === e.element) &&
								coati_array.all(e.specs, node.previousSibling) &&
								checkNode(node.previousSibling, elements));
						}
						return false;
					};

					var foundNodes = [];
					for (var gi = 0; gi < groups.length; gi++)
					{
						// Find the deepest-level element everywhere within the specified root element
						elements = groups[gi];
						var e0 = elements[elements.length - 1];   // "var e" would give a naming conflict and it would get overwritten somewhere else
						var nodes = o.getElementsByTagName(e0.element);   // returns a HTMLCollection...
						nodes = coati_array.filter(nodes,
							function(node) {
								return coati_array.all(e0.specs, node) && checkNode(node, elements);
							}
						);
						foundNodes = foundNodes.concat(nodes);
					}
					return coati_array.uniq(foundNodes);
				},


				// ====================  EVENTS  ====================

				// Attaches an event to an element
				//
				// The event handler will be wrapped in a compatibility layer offering a
				// DOM-compatible event object as the handler's parameter. If the handler
				// returns false, the event will be cancelled.
				//
				// The event handler function will be passed the event object as only argument.
				// The source element can be retrieved through e.target.
				//
				// o = Element
				// n = Event name (DOM syntax, i.e. without "on"). The special name "domload" is called
				//     right after the DOM tree has loaded (in Firefox/Mozilla) and is mapped to "load"
				//     for other browsers.
				// handler = Event handler function
				//
				// Returns a handle to allow detaching the event again.
				//
				addEvent: function(o, n, handler, norec)
				{
					if (!norec)
					{
						// Special case "textchanged" event:
						// The input event only works in Firefox and Opera, the propertychange event only works in IE.
						// So let's just add both events to always know when an input field's text value has changed.
						// It may fire a bit too often, though, but that's simply the nature of it and it can be
						// handled in the calling application.
						if (n == "textchanged")
							return [coati_dom.addEvent(o, "input", handler, true),
								coati_dom.addEvent(o, "propertychange", handler, true)];

						// Spacial case "transitionend" event:
						// Firefox 4 supports it, but Webkit and Opera have different names for it. Register all of them.
						if (n == "transitionend")
							return [coati_dom.addEvent(o, "transitionend", handler, true),
								coati_dom.addEvent(o, "webkitTransitionEnd", handler, true),
								coati_dom.addEvent(o, "oTransitionEnd", handler, true)];
					}

					var wrappedHandler = function(e)
					{
						// Unify event object for DOM/IE
						e = findEvent(e);
						var res;
						//try
						{
							res = handler(e);
						}
						//catch (ex)
						{
						//	coati.errorMessage(coati.formatException(ex), n + " event handler function");
						}
						if (res === false)
						{
							// Cancel event bubbling - "don't tell my parents!"
							e.cancelBubble = true;   // IE
							if (e.stopPropagation)
								e.stopPropagation();   // DOM2

							// Prevent default action (doesn't work in Opera)
							e.returnValue = false;
							//try { e.preventDefault(); } catch (ex) { }   // DOM2
							if (e.preventDefault)
								e.preventDefault();   // TODO: Test
							return false;   // IE
						}
						if (isString(res))
						{
							// This is only useful for the beforeunload event to pass a text message to the user dialog
							e.returnValue = res;   // IE, FF < 4 (FF 4 won't display our message at all - see bug 588292)
							return res;   // Safari
						}
						return true;   // IE
					};

					// Find a compatible event name for some less competent browsers
					var cn = n;
					if (n == "domload")
					{
						n = "DOMContentLoaded";
						cn = "load";
					}

					o = coati_dom_select(o);
					if (!isSet(o)) return;
					if (o.addEventListener)
						o.addEventListener(n, wrappedHandler, false);   // DOM, IE9
					else if (o.attachEvent)
						o.attachEvent("on" + cn, wrappedHandler);   // IE
					else
						o["on" + cn] = wrappedHandler;   // even older - Attention: This removes previously assigned events (?)

					return {"o": o, "n": n, "cn": cn, "f": wrappedHandler };
				},

				// Removes an event from an element
				//
				// h = Return value from previous addEvent call (Arrays accepted)
				//
				removeEvent: function(h)
				{
					if (!isArray(h)) h = [h];
					h = coati_array.flatten(h);
					coati_array_each(h, function(h)
					{
						if (h.o.addEventListener)
							h.o.removeEventListener(h.n, h.f, false);   // DOM
						else if (h.o.attachEvent)
							h.o.detachEvent("on" + h.cn, h.f);   // IE
						else
							h.o["on" + h.cn] = undefined;   // even older - Attention: This can only remove all events (?)
					});
				},

				// Add a handler function to the historyChanged event that fires whenever
				// the user presses the browser's back/forward buttons and a page state has
				// been set using location.hash.
				//
				addHistoryChangedHandler: function(h)
				{
					historyChangedHandlers.push(h);

					if (historyChangedHandlers.length == 1)
					{
						// Just added the first handler: start the whole mechanism

						prevLocationHash = window.location.hash;

						// Monitor the location hash and react on any changes
						// (i.e. when the user presses the browser's back/forward button)
						setInterval(function() {
							if (window.location.hash != prevLocationHash)
							{
								prevLocationHash = window.location.hash;
								coati_dom.onHistoryChanged();
							}
						}, 100);
					}
				},

				// Fires the historyChanged event. For internal use only.
				//
				onHistoryChanged: function()
				{
					coati_array_each(historyChangedHandlers, function(h) { h(); });
				},

				// Executes a function after the page has loaded or immediately if this is already the case
				//
				whenDomLoaded: function(fn)
				{
					if (domLoaded)
						fn();
					else
						coati_dom.addEvent(window, "domload", fn);
				},

				// Has the DOM loaded?
				//
				isDomLoaded: function()
				{
					return domLoaded;
				},

				// For internal use only!
				//
				setDomLoaded: function()
				{
					domLoaded = true;
				},

				// Executes a function after the page has loaded or immediately if this is already the case
				//
				whenPageLoaded: function(fn)
				{
					if (pageLoaded)
						fn();
					else
						coati_dom.addEvent(window, "load", fn);
				},

				// Has the page loaded?
				//
				isPageLoaded: function()
				{
					return pageLoaded;
				},

				// For internal use only!
				//
				setPageLoaded: function()
				{
					pageLoaded = true;
				},


				// ====================  KEYBOARD EVENTS  ====================

				// Array of key names
				//
				keyNames: {
					8: "Backspace",
					9: "Tab",
					13: "Enter",
					16: "Shift",
					17: "Ctrl",
					18: "Alt",
					19: "Pause",
					20: "CapsLock",
					27: "Escape",
					32: "Space",
					33: "PageUp",
					34: "PageDown",
					35: "End",
					36: "Home",
					37: "Left",
					38: "Up",
					39: "Right",
					40: "Down",
					45: "Insert",
					46: "Delete",
					48: "0",
					49: "1",
					50: "2",
					51: "3",
					52: "4",
					53: "5",
					54: "6",
					55: "7",
					56: "8",
					57: "9",
					65: "A",
					66: "B",
					67: "C",
					68: "D",
					69: "E",
					70: "F",
					71: "G",
					72: "H",
					73: "I",
					74: "J",
					75: "K",
					76: "L",
					77: "M",
					78: "N",
					79: "O",
					80: "P",
					81: "Q",
					82: "R",
					83: "S",
					84: "T",
					85: "U",
					86: "V",
					87: "W",
					88: "X",
					89: "Y",
					90: "Z",
					91: "WinLeft",
					92: "WinRight",
					93: "AppMenu",
					96: "Num0",
					97: "Num1",
					98: "Num2",
					99: "Num3",
					100: "Num4",
					101: "Num5",
					102: "Num6",
					103: "Num7",
					104: "Num8",
					105: "Num9",
					106: "Multiply",
					107: "Add",
					109: "Subtract",
					111: "Divide",
					112: "F1",
					113: "F2",
					114: "F3",
					115: "F4",
					116: "F5",
					117: "F6",
					118: "F7",
					119: "F8",
					120: "F9",
					121: "F10",
					122: "F11",
					123: "F12",
					144: "NumLock",
					145: "ScrollLock",
					170: "Search",
					172: "Web",
					180: "Mail"
				},

				// Attachs a keyboard event to an element
				//
				// See DOMEvents.add for notes about the handler function.
				//
				// keyName = (string) Key combination name
				// handler = Event handler function
				//
				addKeyEvent: function(o, keyName, handler)
				{
					return coati_dom.addEvent(o, "keydown", function(e) {
						try
						{
							if (coati_dom.isKeyName(e, keyName))
								return handler(e);
						}
						catch (ex)
						{
							coati.errorMessage("Exception caught while processing the key event handler for " + keyName + ":\n" + coati.formatException(ex));
						}
					});
				},

				// Attaches a keyboard event to the entire window
				//
				// See DOMKeyEvent.add for details
				//
				addGlobalKeyEvent: function(keyName, handler)
				{
					return coati_dom.addKeyEvent(document, keyName, handler);
				},

				// Removes an event from an element
				//
				// h = Return value from previous addEvent call (Arrays accepted)
				//
				removeKeyEvent: function(h)
				{
					coati_dom.removeEvent(h);
				},

				// Converts a key name into its key code
				//
				keyCodeFromName: function(n)
				{
					for (var c in self.keyNames)
						if (self.keyNames[c] == n) return c;
					return 0;
				},

				// Converts a key code into its key name
				//
				keyNameFromCode: function(c)
				{
					if (c in self.keyNames) return self.keyNames[c];
					return "";
				},

				// Was the specified key combination pressed in this event?
				//
				isKeyName: function(e, n)
				{
					var modifier = false;
					var modAlt = false;
					var modCtrl = false;
					var modShift = false;
					do
					{
						modifier = false;
						if (coati_string.startsWith(n, "Alt+"))
						{
							modAlt = true;
							n = n.substr(4);
							modifier = true;
						}
						if (coati_string.startsWith(n, "Ctrl+"))
						{
							modCtrl = true;
							n = n.substr(5);
							modifier = true;
						}
						if (coati_string.startsWith(n, "Shift+"))
						{
							modShift = true;
							n = n.substr(6);
							modifier = true;
						}
					}
					while (modifier);
					if (modAlt ^ e.altKey) return false;
					if (modCtrl ^ e.ctrlKey) return false;
					if (modShift ^ e.shiftKey) return false;
					return findKeyCode(e) == coati_dom.keyCodeFromName(n);
				},

				// Composes a key combination name from the event to display it to the user
				//
				getKeyName: function(e)
				{
					var n = "";
					if (e.ctrlKey) n += "Ctrl+";
					if (e.shiftKey) n += "Shift+";
					var keyCode = findKeyCode(e);
					var singleName = coati_dom.keyNameFromCode(keyCode);
					if (singleName)
						n += singleName;
					else
						n += "<" + keyCode + ">";
					return n;
				},


				// ====================  TEXT EDITING  ====================

				// Inserts text into a textarea element.
				//
				insertText: function(textbox, openText, closeText)
				{
					if (isString(textbox))
						textbox = coati_dom_select(textbox);

					// API check
					if (textbox.selectionStart >= 0)
					{
						var start_selection = textbox.selectionStart;
						var end_selection = textbox.selectionEnd;
						var sel_length = end_selection - start_selection;
						var scroll_top = textbox.scrollTop;
						var scroll_left = textbox.scrollLeft;

						if (isSet(closeText))
						{
							// Exclude whitespace from the start and end
							while (end_selection - start_selection > 1 && textbox.value.substr(start_selection, 1).match(/\s/))
								start_selection--;
							while (end_selection - start_selection > 1 && textbox.value.substr(end_selection - 1, 1).match(/\s/))
								end_selection--;
						}

						// Fetch everything from start of text area to selection start
						var before = textbox.value.substring(0, start_selection);
						// Fetch everything from start of selection to end of selection
						var seltext = textbox.value.substring(start_selection, end_selection);
						// Fetch everything from end of selection to end of text area
						var after = textbox.value.substring(end_selection, textbox.textLength);

						if (sel_length > 0)
						{
							// Have text selection

							if (isSet(closeText))
							{
								if (seltext.substring(0, openText.length) == openText &&
									seltext.substring(seltext.length - closeText.length, seltext.length) == closeText)
								{
									// Opening and closing text is already there, remove it
									seltext = seltext.substring(openText.length, seltext.length - closeText.length);
									end_selection -= openText.length + closeText.length;
								}
								else
								{
									// Enclose selected text in opening and closing text
									seltext = openText + seltext + closeText;
									end_selection += openText.length + closeText.length;
								}
							}
							else
							{
								// Replace selected text
								seltext = openText;
								start_selection = end_selection = start_selection + openText.length;
							}
						}
						else
						{
							// No text selected

							// Just insert the new text
							seltext = openText + (isSet(closeText) ? closeText : "");
							start_selection = end_selection = start_selection + openText.length;
						}

						textbox.value = before + seltext + after;
						textbox.selectionStart = start_selection;
						textbox.selectionEnd = end_selection;
						textbox.focus();
						textbox.scrollTop = scroll_top;
						textbox.scrollLeft = scroll_left;
					}
					else
					{
						// IE < 9 doesn't support this at all
						alert("Your browser does not support this operation.");
					}
				},


				// Puts a string at the beginning of each selected line in a textarea element.
				//
				insertLineStart: function(textbox, lineStart)
				{
					if (isString(textbox))
						textbox = coati_dom_select(textbox);

					// API check
					if (textbox.selectionStart >= 0)
					{
						var start_selection = textbox.selectionStart;
						var end_selection = textbox.selectionEnd;
						var sel_length = end_selection - start_selection;
						var scroll_top = textbox.scrollTop;
						var scroll_left = textbox.scrollLeft;

						// Exclude line breaks from the end
						while (end_selection - start_selection > 1 && textbox.value.substr(end_selection - 1, 1) == "\n")
							end_selection--;

						// Find the beginning of the current line
						for (var i = start_selection; i > 0; i--)
						{
							if (textbox.value.charCodeAt(i) == 10)
							{
								i++;
								break;
							}
						}
						start_selection = i;

						// Process multiple selected lines - find the end of the last line
						for (var j = end_selection; j < textbox.value.length - 1; j++)
						{
							if (textbox.value.charCodeAt(j) == 10)
							{
								break;
							}
						}
						end_selection = j;

						// Fetch everything from start of text area to selection start
						var before = textbox.value.substring(0, start_selection);
						// Fetch everything from start of selection to end of selection
						var middle = textbox.value.substring(start_selection, end_selection);
						// Fetch everything from end of selection to end of text area
						var after = textbox.value.substring(end_selection, textbox.textLength);

						var addedCount = middle.length;
						// Replace all line breaks except the last one
						middle = middle.replace(/\n(?!$)/g, "\n" + lineStart);
						addedCount = middle.length - addedCount;
						addedCount += lineStart.length;   // Add the first line

						textbox.value = before + lineStart + middle + after;

						textbox.selectionStart = start_selection;
						textbox.selectionEnd = end_selection + addedCount;
						textbox.focus();
						textbox.scrollTop = scroll_top;
						textbox.scrollLeft = scroll_left;
					}
					else
					{
						// IE < 9 doesn't support this at all
						alert("Your browser does not support this operation.");
					}
				}
			};

			return self;
		}(),


		// **********************************************************************
		// ***************                                        ***************
		// ***************              UI EFFECTS                ***************
		// ***************                                        ***************
		// **********************************************************************

		fx: function()
		{
			// State variables for image zooming
			var openLayer;
			var openImage;
			var openText;
			var registeredZoomUrls = [];
			
			// Registers the popstate event handler upon first use, to zoom an image again when
			// navigating in history.
			//
			function ensurePopstateHandler()
			{
				if (registeredZoomUrls.length == 1)
				{
					coati_dom.addEvent(window, "popstate", function(e)
					{
						if (e.state && e.state.coati == "image-zoom")
						{
							for (var i = 0; i < registeredZoomUrls.length; i++)
							{
								if (registeredZoomUrls[i].url == e.state.url)
								{
									registeredZoomUrls[i].action(undefined, true);
									break;
								}
							}
						}
					});
				}
			}

			return {
				// Converts the given textarea element to an auto-height element.
				//
				// This re-arranges the elements in the DOM tree and uses advanced CSS styles to
				// have the browser do the (mostly accurate) size calculations. The textarea
				// element is replaced with a container div and moved into it.
				//
				// The idea is based on http://stackoverflow.com/questions/7477/autosizing-textarea/2032642#2032642
				// and was completely put in this JavaScript function for simple plug-in use.
				//
				// Tested with Firefox 4.0, IE 8, Chrome 10 and Opera 11.0.
				//
				// textbox = (element/string) The textarea element to work on
				// minRows = (int) Minimum of visible rows, default: 3
				// maxRows = (int) Maximum of visible rows, default: 15
				//
				makeTextAutoHeight: function(textbox, minRows, maxRows)
				{
					minRows = minRows || 3;
					maxRows = maxRows || 15;

					if (isString(textbox))
						textbox = coati_dom_select(textbox);

					var container = document.createElement("div");
					container.style.position = "relative";
					// Move the original textarea's styles to the container element
					container.style.width = coati_dom.width(textbox) + "px";

					// Replace the original textarea by the new container
					textbox.parentNode.replaceChild(container, textbox);
					// Prepare the textarea's style
					textbox.style.boxSizing = "border-box";
					textbox.style.MozBoxSizing = "border-box";
					textbox.style.MsBoxSizing = "border-box";
					textbox.style.WebkitBoxSizing = "border-box";
					textbox.style.height = "100%";
					textbox.style.position = "absolute";
					textbox.style.width = "100%";
					textbox.style.resize = "none";   // Gecko 2
					// Add the textarea to the container
					container.appendChild(textbox);

					// Min/max rows calculations, may be unavailable in IE
					var lineHeight = coati_dom.getEffectiveStyle(textbox, "line-height");
					if (lineHeight && lineHeight.match(/px/))
						lineHeight = parseInt(lineHeight);
					else
						lineHeight = 0;

					// Create the shadow copy element
					var copy = document.createElement("div");
					copy.style.boxSizing = "border-box";
					copy.style.MozBoxSizing = "border-box";
					copy.style.MsBoxSizing = "border-box";
					copy.style.WebkitBoxSizing = "border-box";
					copy.style.width = "100%";
					copy.style.whiteSpace = "pre-wrap";
					copy.style.visibility = "hidden";
					//copy.style.color = "red";   // Enable for debugging, also disable hidden visibility
					//textbox.style.backgroundColor = "transparent";   // Enable for debugging

					if (lineHeight)
					{
						copy.style.minHeight = (minRows * lineHeight + 1) + "px";
						copy.style.maxHeight = (maxRows * lineHeight + 1) + "px";
					}
					else
					{
						copy.style.minHeight = (minRows * 1.4) + "em";
						copy.style.maxHeight = (maxRows * 1.4) + "em";
					}

					// Add the shadow copy to the container
					container.appendChild(copy);

					// Add event handling
					var copyStyle = function(name, offset)
					{
						var value = coati_dom.getEffectiveStyle(textbox, name);
						if (offset)
							value = parseInt(value) + offset + "px";
						copy.style[coati_string.toCamelCase(name)] = value;
					};

					var autoSize = function()
					{
						// Copy textarea contents; browser will calculate correct height of copy,
						// which will make overall container taller, which will make textarea taller
						var text = textbox.value;

						text = text.replace(/&/g, "&amp;").replace(/</g, "&lt;").replace(/>/g, "&gt;").replace(/\r/g, "");
						var newLine = text.substr(text.length - 1) == "\n";   // IE8 doesn't support -1 here

						if (coati.browser.ie())
						{
							// IE doesn't regard the white-space specification. Oddly it only regards
							// it when it was disabled and enabled again through the developer toolbar.
							// Deferring the setting doesn't help either. So we do it the classic way.
							text = text.replace(/\n/g, "<br />").replace(/  /g, "&nbsp; ").replace(/  /g, "&nbsp; ");
						}

						// Make sure the last line break is visible
						if (newLine)
							text += ".";

						copy.innerHTML = text;

						// Copy all styles for the exact same layout
						copyStyle("padding-top", 1);
						copyStyle("padding-right", 1);
						copyStyle("padding-bottom", 1);
						copyStyle("padding-left", 1);
						copyStyle("border-top-width");
						copyStyle("border-right-width");
						copyStyle("border-bottom-width");
						copyStyle("border-left-width");
						copyStyle("border-top-style");
						copyStyle("border-right-style");
						copyStyle("border-bottom-style");
						copyStyle("border-left-style");
						copyStyle("font-family");
						copyStyle("font-size");
						copyStyle("font-style");
						copyStyle("line-height");
					};

					coati_dom.addEvent(textbox, "textchanged", autoSize);
					autoSize();
				},

				// Makes the image zoomable by clicking on it.
				//
				// img = (element/string) The img element to zoom
				//
				makeImageZoomable: function(img, src, width, height, subtitle)
				{
					var haveText = isSet(subtitle) && subtitle != "";

					// Add click event handler to the image
					img.style.cursor = "pointer";
					coati_dom.addEvent(img, "click", zoomInto);

					registeredZoomUrls.push({ url: src, action: zoomInto });
					ensurePopstateHandler();
					
					var startBorderWidth;
					var startLeft, startTop, startWidth, startHeight;
					var transitionSupported;
					var keyEventHandles = [];

					var maxLayerOpacity = 0.8;

					// Starts the animation to zoom into the image
					//
					// popped: (bool) false/unset to create a new history entry, true if coming from a history entry
					//
					function zoomInto(e, popped)
					{
						// Safety exit
						if (isSet(openImage) || isSet(openText) || isSet(openLayer)) return;

						// Create large image
						var img2 = document.createElement("img");
						openImage = img2;
						coati_dom.addEvent(img2, "load", continueZoomInto);
						coati_dom.addEvent(img2, "error", cancelZoomInto);

						img.style.cursor = "wait";   // Indicate the loading process
						img2.src = src;

						if (history.pushState)
						{
							if (!popped)
							{
								history.pushState({ coati: "image-zoom", url: src }, "");
							}
							var eh = coati_dom.addEvent(window, "popstate", function(e)
							{
								if (!isSet(e.state) || e.state.coati != "image-zoom")
								{
									coati_dom.removeEvent(eh);
									closeZoom();
								}
							});
						}
					}

					// Cancel the process now that the larger image could not be loaded.
					//
					function cancelZoomInto(e)
					{
						img.style.cursor = "pointer";   // Reset cursor
						openImage = undefined;
						alert(coati.getConfigValue("zoomableimage.loaderror", "Error loading this image."));
					}

					// Continues the animation now that the larger image has been loaded.
					//
					// The image must be loaded before the transition is started or it will disturb
					// the transition.
					//
					function continueZoomInto(e)
					{
						img.style.cursor = "pointer";   // Reset cursor

						startBorderWidth = parseInt(coati.dom.getEffectiveStyle(img, "padding-top"));

						var openImageStyle = openImage.style;
						openImageStyle.boxShadow = "0 2px 8px rgba(0, 0, 0, 0.5)";
						openImageStyle.border = "solid " + startBorderWidth + "px #f6f6f6";
						//openImageStyle.cursor = "pointer";

						// Create background layer
						var layer = document.createElement("div");
						var layerStyle = layer.style;
						layerStyle.position = "fixed";
						layerStyle.zIndex = "900";
						layerStyle.top = 0;
						layerStyle.left = 0;
						// Mobile browsers don't support fixed positions, so make the layer large
						// enough to reach down to where the viewport is scrolled to:
						layerStyle.width = "1000%";
						layerStyle.height = "10000%";
						layerStyle.backgroundColor = self.getConfigValue("imagezoom.backcolor", "#808080");
						if (isSet(layerStyle.WebkitTapHighlightColor))
						{
							// Remove tap highlighting when closing the zoom image
							layerStyle.WebkitTapHighlightColor = "rgba(0,0,0,0)";
						}

						coati_dom.setOpacity(layer, 0);
						document.body.appendChild(layer);
						openLayer = layer;

						// Remember original position
						startLeft = coati_dom.left(img);
						startTop = coati_dom.top(img);
						startWidth = coati_dom.width(img) - 2 * startBorderWidth;
						startHeight = coati_dom.height(img) - 2 * startBorderWidth;

						// Move to start position
						openImageStyle.position = "absolute";
						openImageStyle.zIndex = "901";
						openImageStyle.left = startLeft + "px";
						openImageStyle.top = startTop + "px";
						openImageStyle.width = startWidth + "px";
						openImageStyle.height = startHeight + "px";
						//openImageStyle.opacity = "0";   // Don't use IE compatibility for performance reasons (IE9 should improve it but supports this anyway)
						document.body.appendChild(openImage);

						coati_dom.addEvent(openImage, "click", onCloseZoom);
						coati_dom.addEvent(layer, "click", onCloseZoom);

						// Prepare animation
						var endWidth = width;
						var endHeight = height;
						var windowWidth = coati_dom.windowClientWidth();
						var windowHeight = coati_dom.windowClientHeight();
						// Constrain size to the available space in the window
						var hMargin = 10;
						if (windowWidth < 500)
						{
							hMargin = 0;
						}
						var hSpace = 2 * 15 + 2 * hMargin;
						var vSpace = 2 * 15 + 2 * 10 + /*textGap*/ 10 + /*textHeight*/ 35;
						var endWidth = Math.min(width, windowWidth - hSpace);
						var endHeight = endWidth / width * height;
						endHeight = Math.min(endHeight, windowHeight - vSpace);
						endWidth = endHeight / height * width;
						var endLeft = coati_dom.scrollX() + windowWidth / 2 - endWidth / 2 - 15 || 0;
						//if (endWidth >= windowWidth - hSpace)
						//{
						//	endWidth = windowWidth - hSpace;
						//	endLeft = 0;
						//}
						var endTop = coati_dom.scrollY() + windowHeight * 0.47 - endHeight / 2 - 15 || 0;
						if (endTop < 10)
						{
							endTop = 10;
						}

						// Create text
						if (haveText)
						{
							openText = document.createElement("div");
							var openTextStyle = openText.style;
							openTextStyle.visibility = "hidden";
							openTextStyle.position = "absolute";
							openTextStyle.zIndex = "901";
							//openTextStyle.cursor = "pointer";
							openTextStyle.width = endWidth + "px";
							openTextStyle.left = endLeft + 15 + "px";
							openTextStyle.top = endTop + 15 + endHeight + 10 + "px";
							openText.innerHTML = subtitle;
							document.body.appendChild(openText);

							coati_dom.addEvent(openText, "click", onCloseZoom);
						}

						var transDuration = "0.3s";
						var imgTrans = "left " + transDuration + ", top " + transDuration + ", width " + transDuration + ", height " + transDuration + ", border-width " + transDuration + ", opacity " + transDuration;
						var layerTrans = "opacity " + transDuration;

						var transitionName = coati.browser.transition();
						if (transitionName)
						{
							openImageStyle[transitionName] = imgTrans;
							layerStyle[transitionName] = layerTrans;
							transitionSupported = true;
						}

						keyEventHandles.push(coati_dom.addGlobalKeyEvent("Escape", onCloseZoom));

						var eventHandle;

						var finished = function()
						{
							// Clean up, this function must not be called after the closing transition anymore!
							if (eventHandle)
								coati_dom.removeEvent(eventHandle);

							if (haveText)
								openTextStyle.visibility = "";
						};

						if (transitionSupported)
						{
							// Give it some time to apply the transition properties or it won't animate.
							// 20ms is enough, too, so 50ms should be pretty safe.
							coati.defer(function() {
								openImageStyle.left = endLeft + "px";
								openImageStyle.top = endTop + "px";
								openImageStyle.width = endWidth + "px";
								openImageStyle.height = endHeight + "px";
								openImageStyle.borderWidth = "15px";
								if (haveText)
									openImageStyle.borderBottomWidth = 10 + coati_dom.height(openText) + 15 + "px";
								//openImageStyle.opacity = "1";
								layerStyle.opacity = maxLayerOpacity;
							}, 50);

							eventHandle = coati_dom.addEvent(openImage, "transitionend", finished);
						}
						else
						{
							var duration = 150;   // Try faster, it will be slow anyway

							new coati.animation({o: openImage, type: "left", end: endLeft, duration: duration, play: true});
							new coati.animation({o: openImage, type: "top", end: endTop, duration: duration, play: true});
							new coati.animation({o: openImage, type: "width", end: endWidth, duration: duration, play: true});
							new coati.animation({o: openImage, type: "height", end: endHeight, duration: duration, play: true, onfinished: finished});
							if (coati.browser.ie() && coati.browser.ieVersion() < 9)
							{
								// Lo-fi solution
								coati_dom.setOpacity(layer, maxLayerOpacity);
							}
							else
							{
								new coati.animation({o: layer, type: "opacity", end: maxLayerOpacity, duration: duration, play: true});
							}
							openImageStyle.borderWidth = "15px";
							if (haveText)
								openImageStyle.borderBottomWidth = 10 + coati_dom.height(openText) + 15 + "px";
						}
					}

					// Event handler to start closing the image zoom.
					//
					function onCloseZoom(e)
					{
						if (history.pushState)
						{
							history.back();
						}
						else
						{
							closeZoom();
						}
					}
					
					// Starts the closing animation.
					//
					function closeZoom(e)
					{
						coati_dom.removeEvent(keyEventHandles);

						if (isSet(openText))
						{
							openText.parentNode.removeChild(openText);
							openText = undefined;
						}

						if (isSet(openImage) && isSet(openLayer))
						{
							if (transitionSupported)
							{
								var openImageStyle = openImage.style;
								openImageStyle.left = startLeft + "px";
								openImageStyle.top = startTop + "px";
								openImageStyle.width = startWidth + "px";
								openImageStyle.height = startHeight + "px";
								//openImageStyle.borderWidth = "0px";
								openImageStyle.borderWidth = startBorderWidth + "px";
								//openImageStyle.opacity = "0";
								openLayer.style.opacity = 0;

								coati_dom.addEvent(openImage, "transitionend", function(e) {
									if (openImage && openImage.parentNode)
									{
										openImage.parentNode.removeChild(openImage);
										openImage = undefined;
									}
									if (openLayer && openLayer.parentNode)
									{
										openLayer.parentNode.removeChild(openLayer);
										openLayer = undefined;
									}
								});
							}
							else
							{
								var duration = 150;   // Try faster, it will be slow anyway

								new coati.animation({o: openLayer, type: "opacity", start: maxLayerOpacity, end: 0, duration: duration, play: true});
								new coati.animation({o: openImage, type: "left", end: startLeft, duration: duration, play: true});
								new coati.animation({o: openImage, type: "top", end: startTop, duration: duration, play: true});
								new coati.animation({o: openImage, type: "width", end: startWidth, duration: duration, play: true});
								new coati.animation({o: openImage, type: "height", end: startHeight, duration: duration, play: true, onfinished: function() {
									openImage.parentNode.removeChild(openImage);
									openImage = undefined;
									openLayer.parentNode.removeChild(openLayer);
									openLayer = undefined;
								}});
							}
						}
					}
				},

				// Starts an image slideshow.
				//
				// div = (element/string) The element to run the slideshow in.
				// delay = (int) Seconds between slides.
				//
				slideshow: function(div, delay)
				{
					var transitionName = coati.browser.transition();
					var items = [];
					var currentIndex = 0;
					var nextIndex;
					var eventHandle;
					if (!isSet(delay)) delay = 6;

					div = $(div);

					//transitionName = undefined;   // TEST

					for (var i = 0; i < div.childNodes.length; i++)
					{
						var divChild = div.childNodes[i];
						if (divChild.nodeType == 1 /*element*/ && divChild.nodeName.toLowerCase() == "div")
						{
							//var item = divChild.childNodes[0];
							//while (isSet(item) && item.nodeType != 1)
							//{
							//	item = item.nextSibling;
							//}
							//if (!isSet(item)) return;   // Ignore items with no element in them

							var item = divChild;
							if (items.length > 0)
							{
								coati_dom.hide(item);
								coati_dom.setOpacity(item, 0);
							}
							//divChild.style.marginBottom = -coati_dom.height(div) + "px";
							if (transitionName)
							{
								item.style[transitionName] = "opacity 1s";
							}
							items.push(item);   // Collect all slideshow items
						}
					}

					setTimeout(nextItem, delay * 1000);

					function nextItem()
					{
						nextIndex = (currentIndex + 1) % items.length;
						coati_dom.show(items[nextIndex]);

						if (transitionName)
						{
							// Give it some time to apply the transition properties or it won't animate.
							// 20ms is enough, too, so 50ms should be pretty safe.
							coati.defer(function() {
								items[currentIndex].style.opacity = 0;
								items[nextIndex].style.opacity = 1;
							}, 50);

							eventHandle = coati_dom.addEvent(items[nextIndex], "transitionend", scheduleNext);
							// WORKAROUND:
							// The above doesn't work in a background tab in Firefox 6, so we need a safety net for it.
							// See also https://bugzilla.mozilla.org/show_bug.cgi?id=683696
							setTimeout(recoverScheduleNext, 1500);
						}
						else
						{
							new coati.animation({o: items[currentIndex], type: "opacity", start: 1, end: 0, play: true});
							new coati.animation({o: items[nextIndex], type: "opacity", start: 0, end: 1, play: true, onfinished: scheduleNext});
						}
					}

					function scheduleNext()
					{
						//alert("scheduleNext");   // TEST: Does this get called, even in the background tab?
						if (eventHandle)
						{
							coati_dom.removeEvent(eventHandle);
							eventHandle = undefined;
						}
						coati_dom.hide(items[currentIndex]);
						currentIndex = nextIndex;
						setTimeout(nextItem, delay * 1000);
					}

					function recoverScheduleNext()
					{
						if (eventHandle)
							scheduleNext();
					}
				}
			};
		}(),


		// **********************************************************************
		// ***************                                        ***************
		// ***************           AJAX WEB REQUESTS            ***************
		// ***************                                        ***************
		// **********************************************************************

		webrequest: function()
		{
			return {
				// Start an asynchronous HTTP request
				//
				// The server will receive all set cookies as with a regular HTTP web page request.
				// The event handler function is only called upon successful finishing the HTTP request.
				// An error message is displayed otherwise.
				// The event handler function will be passed the request object as its first parameter.
				//
				// url = (string) URL to request
				// method = (string) HTTP method: "GET" or "POST"
				// data = (string) Readily encoded HTTP POST data.
				//        (array) Associative array (actually an object) with all parameters to send in a POST request
				// eventHandler = Function to be called when the request has finished.
				// errorHandler = Function to be called when a server error occured. If unset, a default error message is displayed.
				//
				asyncRequest: function(url, method, data, eventHandler, errorHandler)
				{
					// internal function to create a wrapped event handler
					//
					function makeRequestHandler(request, eventHandler, url, errorHandler)
					{
						if (!isFunction(eventHandler))
						{
							coati.errorMessage("parameter eventHandler is not a function", "asyncRequest, makeRequestHandler");
							return false;
						}
						if (isSet(errorHandler) && !isFunction(errorHandler))
						{
							coati.errorMessage("parameter errorHandler is not a function", "asyncRequest, makeRequestHandler");
							return false;
						}

						return function()
						{
							if (request.readyState == 4)
							{
								if (request.status == 200)
								{
									// do something with the response
									try
									{
										eventHandler(request);
									}
									catch (ex)
									{
										coati.errorMessage("Exception caught while processing the server response:\n" + coati.formatException(ex));
									}
								}
								else
								{
									if (isSet(errorHandler))
									{
										// do something with the response
										try
										{
											errorHandler(request);
										}
										catch (ex)
										{
											coati.errorMessage("Exception caught while processing the server error response:\n" + coati.formatException(ex));
										}
									}
									else
									{
										coati.errorMessage("Server responded with error code " + request.status + " (" + request.statusText + ")\n" +
											"URL: " + /*coati.getURLDirectoryName(document.location.href) +*/ url);
									}
								}
							}
						}
					}

					// Check all parameters
					method = method.toUpperCase();
					if (method != "GET" && method != "POST")
					{
						coati.errorMessage("parameter method is invalid: " + method, "asyncRequest");
						return false;
					}
					var dataToSend = "";
					if (isObject(data))
						for (var property in data)
						{
							if (isBoolean(data[property]) || isNumber(data[property]) || isString(data[property]))
								dataToSend += (dataToSend != "" ? "&" : "") +
									property + "=" + encodeURIComponent(data[property]);
							else if (isArray(data[property]))
								coati_array_each(data[property], function(o)
								{
									dataToSend += (dataToSend != "" ? "&" : "") +
										property + "[]=" + encodeURIComponent(o);
								});
						}
					else if (isString(data))
						dataToSend = data;
					else if (!isSet(data))
						if (method == "POST")
						{
							coati.errorMessage("parameter data is unset but POST method was requested", "asyncRequest");
							return false;
						}
					else
					{
						coati.errorMessage("parameter data is invalid", "asyncRequest");
						return false;
					}

					// Find browser support
					var request = false;
					if (window.XMLHttpRequest)
						try { request = new XMLHttpRequest(); }
						catch (ex) {}
					else if (window.ActiveXObject)
						try { request = new ActiveXObject("Msxml2.XMLHTTP"); }
						catch (ex)
						{
							try { request = new ActiveXObject("Microsoft.XMLHTTP"); }
							catch (ex2) {}
						}

					// Start HTTP request
					if (!request)
					{
						coati.errorMessage("Browser does not support any kind of XMLHttpRequest object", "asyncRequest");
						return false;
					}
					var handler = makeRequestHandler(request, eventHandler, url, errorHandler);
					if (handler == false)
					{
						coati.errorMessage("handler factory did not return a function", "asyncRequest");
						return false;
					}
					request.open(method, url, true);
					request.onreadystatechange = handler;
					if (dataToSend != "")
						request.setRequestHeader('Content-Type', 'application/x-www-form-urlencoded');
					request.send(dataToSend);
					return true;
				},

				asyncGet: function(url, eventHandler, errorHandler)
				{
					return coati.webrequest.asyncRequest(url, "GET", null, eventHandler, errorHandler);
				},

				asyncPost: function(url, data, eventHandler, errorHandler)
				{
					return coati.webrequest.asyncRequest(url, "POST", data, eventHandler, errorHandler);
				},

				// Data parsing functions
				// Hint: Get the server's XML response with request.responseXML

				getErrorType: function(xml)
				{
					if (xml == null) return undefined;
					var n = xml.getElementsByTagName("error")[0];
					if (n) return n.getAttribute("type");
				},

				getErrorValue: function(xml)
				{
					return coati.webrequest.getValue(xml, "error");
				},

				getResultValue: function(xml)
				{
					return coati.webrequest.getValue(xml, "result");
				},

				getValue: function(xml, name, forceArray, onlyFirst)
				{
					if (xml == null) return undefined;
					if (!isSet(forceArray)) forceArray = false;
					if (!isSet(onlyFirst)) onlyFirst = false;

					var n = xml.getElementsByTagName(name);
					if (n.length == 1 && !forceArray || n.length > 0 && onlyFirst)
					{
						return n[0].firstChild ? n[0].firstChild.nodeValue : "";
					}
					else if (n.length > 0 || forceArray)
					{
						var a = new Array();
						for (var i = 0; i < n.length; i++)
						{
							a.push(n[i].firstChild ? n[i].firstChild.nodeValue : "");
						}
						return a;
					}
				}
			};
		}(),


		// **********************************************************************
		// ***************                                        ***************
		// ***************           COLOR CONVERSIONS            ***************
		// ***************                                        ***************
		// **********************************************************************

		color: function()
		{
			// HTML named colours
			var namedColors = {
				aliceblue: "#f0f8ff",
				antiquewhite: "#faebd7",
				aqua: "#00ffff",
				aquamarine: "#7fffd4",
				azure: "#f0ffff",
				beige: "#f5f5dc",
				bisque: "#ffe4c4",
				black: "#000000",
				blanchedalmond: "#ffebcd",
				blue: "#0000ff",
				blueviolet: "#8a2be2",
				brown: "#a52a2a",
				burlywood: "#deb887",
				cadetblue: "#5f9ea0",
				chartreuse: "#7fff00",
				chocolate: "#d2691e",
				coral: "#ff7f50",
				cornflowerblue: "#6495ed",
				cornsilk: "#fff8dc",
				crimson: "#dc143c",
				cyan: "#00ffff",
				darkblue: "#00008b",
				darkcyan: "#008b8b",
				darkgoldenrod: "#b8860b",
				darkgray: "#a9a9a9",
				darkgreen: "#006400",
				darkgrey: "#a9a9a9",
				darkkhaki: "#bdb76b",
				darkmagenta: "#8b008b",
				darkolivegreen: "#556b2f",
				darkorange: "#ff8c00",
				darkorchid: "#9932cc",
				darkred: "#8b0000",
				darksalmon: "#e9967a",
				darkseagreen: "#8fbc8f",
				darkslateblue: "#483d8b",
				darkslategray: "#2f4f4f",
				darkslategrey: "#2f4f4f",
				darkturquoise: "#00ced1",
				darkviolet: "#9400d3",
				deeppink: "#ff1493",
				deepskyblue: "#00bfff",
				dimgray: "#696969",
				dimgrey: "#696969",
				dodgerblue: "#1e90ff",
				firebrick: "#b22222",
				floralwhite: "#fffaf0",
				forestgreen: "#228b22",
				fuchsia: "#ff00ff",
				gainsboro: "#dcdcdc",
				ghostwhite: "#f8f8ff",
				gold: "#ffd700",
				goldenrod: "#daa520",
				gray: "#808080",
				green: "#008000",
				greenyellow: "#adff2f",
				grey: "#808080",
				honeydew: "#f0fff0",
				hotpink: "#ff69b4",
				indianred: "#cd5c5c",
				indigo: "#4b0082",
				ivory: "#fffff0",
				khaki: "#f0e68c",
				lavender: "#e6e6fa",
				lavenderblush: "#fff0f5",
				lawngreen: "#7cfc00",
				lemonchiffon: "#fffacd",
				lightblue: "#add8e6",
				lightcoral: "#f08080",
				lightcyan: "#e0ffff",
				lightgoldenrodyellow: "#fafad2",
				lightgray: "#d3d3d3",
				lightgreen: "#90ee90",
				lightgrey: "#d3d3d3",
				lightpink: "#ffb6c1",
				lightsalmon: "#ffa07a",
				lightseagreen: "#20b2aa",
				lightskyblue: "#87cefa",
				lightslategray: "#778899",
				lightslategrey: "#778899",
				lightsteelblue: "#b0c4de",
				lightyellow: "#ffffe0",
				lime: "#00ff00",
				limegreen: "#32cd32",
				linen: "#faf0e6",
				magenta: "#ff00ff",
				maroon: "#800000",
				mediumaquamarine: "#66cdaa",
				mediumblue: "#0000cd",
				mediumorchid: "#ba55d3",
				mediumpurple: "#9370db",
				mediumseagreen: "#3cb371",
				mediumslateblue: "#7b68ee",
				mediumspringgreen: "#00fa9a",
				mediumturquoise: "#48d1cc",
				mediumvioletred: "#c71585",
				midnightblue: "#191970",
				mintcream: "#f5fffa",
				mistyrose: "#ffe4e1",
				moccasin: "#ffe4b5",
				navajowhite: "#ffdead",
				navy: "#000080",
				oldlace: "#fdf5e6",
				olive: "#808000",
				olivedrab: "#6b8e23",
				orange: "#ffa500",
				orangered: "#ff4500",
				orchid: "#da70d6",
				palegoldenrod: "#eee8aa",
				palegreen: "#98fb98",
				paleturquoise: "#afeeee",
				palevioletred: "#db7093",
				papayawhip: "#ffefd5",
				peachpuff: "#ffdab9",
				peru: "#cd853f",
				pink: "#ffc0cb",
				plum: "#dda0dd",
				powderblue: "#b0e0e6",
				purple: "#800080",
				red: "#ff0000",
				rosybrown: "#bc8f8f",
				royalblue: "#4169e1",
				saddlebrown: "#8b4513",
				salmon: "#fa8072",
				sandybrown: "#f4a460",
				seagreen: "#2e8b57",
				seashell: "#fff5ee",
				sienna: "#a0522d",
				silver: "#c0c0c0",
				skyblue: "#87ceeb",
				slateblue: "#6a5acd",
				slategray: "#708090",
				slategrey: "#708090",
				snow: "#fffafa",
				springgreen: "#00ff7f",
				steelblue: "#4682b4",
				tan: "#d2b48c",
				teal: "#008080",
				thistle: "#d8bfd8",
				tomato: "#ff6347",
				turquoise: "#40e0d0",
				violet: "#ee82ee",
				wheat: "#f5deb3",
				white: "#ffffff",
				whitesmoke: "#f5f5f5",
				yellow: "#ffff00",
				yellowgreen: "#9acd32"
			};

			return {
				// ====================  NOTATION CONVERSION  ====================

				// Converts any known notation to an associative array/object of {R, G, B}. This includes about
				// anything the CSS specs up to version 3 define, plus some extras.
				//
				// x = a colour specification in any known format
				// r = (bool) true if the colour component values shall be rounded to integers
				//
				parse: function(x, r)
				{
					// Internal object structure
					if (isObject(x) && isNumber(x.R) && isNumber(x.G) && isNumber(x.B))
					{
						if (r)
							return {R: Math.round(x.R), G: Math.round(x.G), B: Math.round(x.B)};
						return x;
					}

					if (isString(x))
					{
						var m;
						// HTML long notation
						if (m = x.match(/^#?([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})$/i))
							return {R: parseInt(m[1], 16), G: parseInt(m[2], 16), B: parseInt(m[3], 16)};
						// HTML short notation
						if (m = x.match(/^#?([0-9a-f])([0-9a-f])([0-9a-f])$/i))
							return {R: parseInt(m[1] + m[1], 16), G: parseInt(m[2] + m[2], 16), B: parseInt(m[3] + m[3], 16)};
						// CSS rgb notation with percentage
						if (m = x.match(/^rgb\s*\(\s*([0-9]{1,3})\s*%\s*,\s*([0-9]{1,3})\s*%\s*,\s*([0-9]{1,3})\s*%\s*\)$/i))
							return coati.color.parse({R: +m[1] * 2.55, G: +m[2] * 2.55, B: +m[3] * 2.55}, r);   // Try again for rounding
						// CSS rgb notation
						if (m = x.match(/^rgb\s*\(\s*([0-9]{1,3})\s*,\s*([0-9]{1,3})\s*,\s*([0-9]{1,3})\s*\)$/i))
							return {R: +m[1], G: +m[2], B: +m[3]};
						// CSS hsl notation
						if (m = x.match(/^hsl\s*\(\s*([0-9]{1,3})\s*,\s*([0-9]{1,3})\s*%\s*,\s*([0-9]{1,3})\s*%\s*\)$/i))
							return coati.color.hsl2rgb({H: +m[1], S: +m[2] / 100, L: +m[3] / 100});
						// HTML/SVG colour name
						if (namedColors[x.toLowerCase()])
							return coati.color.parse(namedColors[x.toLowerCase()]);   // Try again with the assigned HEX notation
					}

					if (isNumber(x) && x >= 0 && x <= 0xFFFFFF)
					{
						return {R: x & 0xFF, G: (x & 0xFF00) >> 8, B: (x & 0xFF0000) >> 16};
					}
				},

				// TODO
				systemColor: function(n)
				{
					// Create an element
					// Assign the colour by its name
					// Get the computed style/colour
				},

				// Gets the HTML/SVG colour name for a given colour
				//
				// See http://www.w3.org/TR/2003/CR-css3-color-20030514/#svg-color
				//
				name: function(c)
				{
					c = coati.color.toHtml(c);
					if (isSet(c))
						return coati_array.findValue(namedColors, c);
				},

				// Converts a colour from any format to CSS notation: rgb(1, 2, 3)
				//
				toCss: function(c)
				{
					c = coati.color.parse(c, true);
					if (isSet(c))
						return "rgb(" + c.R + ", " + c.G + ", " + c.B + ")";
				},

				// Converts a colour from any format to HTML notation: #010203
				//
				toHtml: function(c)
				{
					c = coati.color.parse(c, true);
					if (isSet(c))
						return ("#" +
							Math.min(Math.max(c.R, 0), 255).toString(16).pad(2, "0", true) +
							Math.min(Math.max(c.G, 0), 255).toString(16).pad(2, "0", true) +
							Math.min(Math.max(c.B, 0), 255).toString(16).pad(2, "0", true)).toLowerCase();
				},

				// Converts a colour from any format to an integer number: 0x010203
				//
				toInt: function(c)
				{
					c = coati.color.parse(c, true);
					if (isSet(c))
						return (c.B << 16) + (c.G << 8) + c.R;
				},


				// ====================  COLOUR SPACE CONVERSION  ====================

				// Converts a colour from RGB to HSL values.
				//
				// The input colour must be {R: 0..255, G: 0..255, B: 0..255}. Any known format is accepted, too.
				// The returned colour is {H: 0..360, S: 0..1, L: 0..1}.
				//
				// See http://en.wikipedia.org/wiki/HSL_color_space#Formal_specifications for the maths
				//
				rgb2hsl: function(rgb)
				{
					rgb = coati.color.parse(rgb);
					if (!isSet(rgb)) return undefined;

					var hsl = {};
					var r = rgb.R / 255;
					var g = rgb.G / 255;
					var b = rgb.B / 255;
					var min = Math.min(r, g, b);
					var max = Math.max(r, g, b);

					hsl.L = (max + min) / 2;

					if (max == min)
						hsl.H = 0;
					else if (max == r)
						hsl.H = (60 * (g - b) / (max - min)) % 360;
					else if (max == g)
						hsl.H = (60 * (b - r) / (max - min) + 120) % 360;
					else if (max == b)
						hsl.H = (60 * (r - g) / (max - min) + 240) % 360;
					if (hsl.H < 0)
						hsl.H += 360;

					if (max == min)
						hsl.S = 0;
					else if (hsl.L <= 0.5)
						hsl.S = (max - min) / (2 * hsl.L);
					else
						hsl.S = (max - min) / (2 - 2 * hsl.L);

					return hsl;
				},

				// Converts a colour from HSL to RGB values.
				//
				// The input colour must be {H: 0..360, S: 0..1, L: 0..1}.
				// The returned colour is {R: 0..255, G: 0..255, B: 0..255}.
				//
				// See http://en.wikipedia.org/wiki/HSL_color_space#Formal_specifications for the maths
				//
				hsl2rgb: function(hsl)
				{
					var rgb = {};
					if (hsl.L < 0.5)
						var q = hsl.L * (1 + hsl.S);
					else
						var q = hsl.L + hsl.S - hsl.L * hsl.S;
					var p = 2 * hsl.L - q;
					var h = hsl.H / 360;   // Normalise hue to 0..1
					var t = {R: h + 1/3, G: h, B: h - 1/3};

					coati_array_each(["R", "G", "B"], function(C) {
						if (t[C] < 0) t[C]++;
						if (t[C] > 1) t[C]--;
						if (t[C] < 1/6)
							rgb[C] = p + ((q - p) * 6 * t[C]);
						else if (t[C] < 1/2)
							rgb[C] = q;
						else if (t[C] < 2/3)
							rgb[C] = p + ((q - p) * 6 * (2/3 - t[C]));
						else
							rgb[C] = p;
						rgb[C] *= 255;
					});
					return rgb;
				},

				isHsl: function(x)
				{
					return isObject(x) && isNumber(x.H) && isNumber(x.S) && isNumber(x.L);
				},


				// ====================  COLOUR MATHS  ====================

				// Blends two colours with a given ratio
				//
				// Input is accepted as HSL or any known RGB format.
				//
				// ratio = (float) Value from 0 to 1 (e.g. 0.5 blends both colours equally)
				//
				// Returns HSL or RGB respectively (like the first colour).
				//
				blend: function(c1, c2, ratio)
				{
					var wasHsl1 = coati.color.isHsl(c1);
					if (wasHsl1)
						c1 = coati.color.hsl2rgb(c1);
					if (coati.color.isHsl(c2))
						c2 = coati.color.hsl2rgb(c2);

					c1 = coati.color.parse(c1);
					c2 = coati.color.parse(c2);
					ratio = Math.min(Math.max(ratio, 0), 1);
					if (isSet(c1) && isSet(c2))
					{
						var c = {
							R: c1.R + (c2.R - c1.R) * ratio,
							G: c1.G + (c2.G - c1.G) * ratio,
							B: c1.B + (c2.B - c1.B) * ratio};

						return wasHsl1 ? coati.color.rgb2hsl(c) : c;
					}
				},

				// Computes the complementary colour.
				//
				// Input is accepted as HSL or any known RGB format.
				//
				// Returns HSL or RGB respectively.
				//
				// TODO: Make it create more than one colour? See Corel Draw 14
				//
				complementary: function(c)
				{
					var wasHsl = coati.color.isHsl(c);
					if (!wasHsl)
						c = coati.color.rgb2hsl(c);

					c.H = (c.H + 180) % 360;

					return !wasHsl ? coati.color.hsl2rgb(c) : c;
				},

				// Computes a lighter colour.
				//
				// Input is accepted as HSL or any known RGB format.
				//
				// l = (number) Range from 0 to 1: 0 does not lighten the colour, 1 makes it white, 0.5 goes halfway there
				//
				// Returns HSL or RGB respectively.
				//
				lighten: function(c, l)
				{
					var wasHsl = coati.color.isHsl(c);
					if (!wasHsl)
						c = coati.color.rgb2hsl(c);

					c.L += l * (1 - c.L);

					return !wasHsl ? coati.color.hsl2rgb(c) : c;
				},

				// Computes a darker colour.
				//
				// Input is accepted as HSL or any known RGB format.
				//
				// l = (number) Range from 0 to 1: 0 does not darken the colour, 1 makes it black, 0.5 goes halfway there
				//
				// Returns HSL or RGB respectively.
				//
				darken: function(c, l)
				{
					var wasHsl = coati.color.isHsl(c);
					if (!wasHsl)
						c = coati.color.rgb2hsl(c);

					c.L *= 1 - l;

					return !wasHsl ? coati.color.hsl2rgb(c) : c;
				},

				// Converts a colour to its grey equivalent
				//
				// Input is accepted as HSL or any known RGB format.
				//
				// Returns HSL or RGB respectively.
				//
				// See http://labs.hellokeita.com/2008/01/19/grayscale-converter/ for the maths
				//
				grey: function(c)
				{
					var wasHsl = coati.color.isHsl(c);
					if (wasHsl)
						c = coati.color.hsl2rgb(c);

					c = coati.color.parse(c, true);
					if (isSet(c))
					{
						var g = c.R * 0.3 + c.G * 0.59 + c.B * 0.11;
						c = {R: g, G: g, B: g};

						return wasHsl ? coati.color.rgb2hsl(c) : c;
					}
				}

			};
		}(),


		// **********************************************************************
		// ***************                                        ***************
		// ***************               ANIMATIONS               ***************
		// ***************                                        ***************
		// **********************************************************************

		// Initialises the animation
		//
		// The only parameter is an object with the animation settings.
		// The following keys are used:
		//   o: Object to animate. Array of nodes or non-unique CSS selectors are okay, then the
		//      start value is determined from the first node, if unspecified.
		//   type: Animation type (see list below)
		//   start: Start value. If undefined, the current value is determined and used.
		//   end: End value
		//   rel: Specifies whether the end value shall be treated relative to the start value.
		//        Unsupported for colours.
		//   duration: Total duration of the animation in milliseconds. Defaults to 1000.
		//   method: Animation method (see list below). Defaults to "sexp".
		//   param: Animation method parameter. Defaults to 1 (for "sexp").
		//   fps: Animation smoothness in frames per second
		//   play: (bool) Autoplay
		//   onfinished: (function) Finished event handler
		//   onstep: (function) Step event handler
		//   next: (Animation) Next animation to start when this one is finished (shortcut for
		//         onfinished = next.play)
		//
		// Supported animation types:
		//   width, height, left, top, right, bottom, {max,min}-{height,width}
		//   {margin,padding}-{bottom, left, right, top}
		//   opacity
		//   font-size, letter-spacing, line-height
		//   color, background-color
		//   scroll-{h,v}
		//
		// Supported animation methods:
		//   linear: Linear movement (fallback for all unknown method names)
		//   sine: Uses a sine function (no further configuration)
		//   exp: Exponential movement
		//        negative values reverse the acceleration
		//        0: no acceleration (linear movement)
		//        acceleration is 2^v
		//        > 0: slow start, fast end
		//        < 0: fast start, slow end
		//   sexp: Symmetric EXPonential
		//         0: no acceleration (linear movement)
		//         > 0: slow start/end, fast in the middle; 1 -> x^2; 2 -> x^4
		//         < 0: fast start/end, slow in the middle; -1 -> x^0.5; -2 -> x^0.25
		//
		animation: function(params)
		{
			// Private variables

			// Contains all settings' names that were specified explicitly for this instance
			var ownSpecs = [];
			// List of Animations attached to this instance
			var attachedAnims = [];

			// The element(s) subject to the animation
			var o;
			// Animation type, see description above
			var animType = "";
			var startVal;
			var endVal;
			var endRelative = false;
			var duration = 1000;
			var fps = 40;
			var method = "sexp";
			var methodParam = 1;
			var autoPlay = false;
			// Separate attached animations, see initialisation below
			var additionalSpecs = [];

			// Runtime variables
			var timePassed = 0;
			var timeoutHandle;
			var playing = false;
			var progress = 0;

			// Private methods

			// Converts arbitrary CSS size values into pixels.
			//
			function cssToPx(s)
			{
				var m = s.match(/([0-9.]+)(px|pt)/);
				if (m) switch (m[2])
				{
					case "px": return +m[1];
				}
				if (s === "normal") return 0;   // TODO: This is wrong, e.g. for "line-height"
			}

			// Determines the start value, if none was specified, from the object to animate.
			//
			// This is called again when endVal was changed e.g. through setDefaults when this instance
			// was attached to another one.
			//
			function findStartEnd()
			{
				var _o = o;
				if (isArray(_o))
					_o = _o[0];
				if (!isSet(startVal) && isSet(_o))   // Relative movement, find current and calculate end value
				{
					switch (animType)
					{
						case "height": startVal = coati_dom.height(_o); break;
						case "left": startVal = coati_dom.left(_o); break;
						case "opacity": startVal = coati_dom.getOpacity(_o); break;
						case "scroll-h": startVal = o.scrollLeft; break;
						case "scroll-v": startVal = o.scrollTop; break;
						case "top": startVal = coati_dom.top(_o); break;
						case "width": startVal = coati_dom.width(_o); break;

						case "background-color":
						case "border-bottom-color":
						case "border-left-color":
						case "border-right-color":
						case "border-top-color":
						case "color":
							startVal = coati_dom.getEffectiveStyle(_o, animType); break;

						case "border-bottom-width":
						case "border-left-width":
						case "border-right-width":
						case "border-top-width":
						case "bottom":
						case "font-size":
						case "letter-spacing":
						case "line-height":
						case "margin-bottom":
						case "margin-left":
						case "margin-right":
						case "margin-top":
						case "max-height":
						case "max-width":
						case "min-height":
						case "min-width":
						case "padding-bottom":
						case "padding-left":
						case "padding-right":
						case "padding-top":
						case "right":
							 startVal = cssToPx(coati_dom.getEffectiveStyle(_o, animType)); break;
					}
				}
				if (!coati_array_contains(animType, "color") && endRelative)
					endVal += startVal;   // End value is relative to dynamic start value

				if (coati_array_contains(animType, "color"))
				{
					startVal = coati.color.parse(startVal);
					endVal = coati.color.parse(endVal);
				}
			}

			// Does the actual value calculation and applies it to the object being animated. (Internal use)
			//
			function doSet()
			{
				var pos = 0;
				var time = timePassed / duration;
				if (method === "sine")
				{
					// we use the sine function from 3pi/2 to 5pi/2
					// scale down linear time percentage from 0...1 to 3pi/2 to 5pi/2
					pos = Math.sin(1.5 * Math.PI + time * Math.PI);
					// translate sine output from -1...1 to 0...1
					pos = (pos + 1) / 2;
				}
				else if (method === "exp")
				{
					pos = Math.pow(time, Math.pow(2, methodParam));
				}
				else if (method === "sexp")   // Symmetric EXPonential
				{
					var exp = Math.pow(2, methodParam);
					if (time <= 0.5)
						pos = Math.pow(time * 2, exp) / 2;
					else
						pos = 1 - Math.pow(1 - (time - 0.5) * 2, exp) / 2;
				}
				else
				{
					pos = time;
				}

				var v;
				if (coati_array_contains(animType, "color"))
					v = coati.color.blend(startVal, endVal, pos);
				else if (animType === "opacity")
					v = startVal + (endVal - startVal) * pos;
				else
					v = Math.round(startVal + (endVal - startVal) * pos);

				var a = isArray(o) ? o : [o];
				coati_array_each(a, function(o) {
					switch (animType)
					{
						case "opacity": coati_dom.setOpacity(o, v); break;
						case "scroll-h": o.scrollLeft = v; break;
						case "scroll-v": o.scrollTop = v; break;

						case "background-color":
						case "border-bottom-color":
						case "border-left-color":
						case "border-right-color":
						case "border-top-color":
						case "color":
							o.style[coati_string.toCamelCase(animType)] = coati.color.toCss(v); break;

						case "border-bottom-width":
						case "border-left-width":
						case "border-right-width":
						case "border-top-width":
						case "bottom":
						case "font-size":
						case "height":
						case "left":
						case "letter-spacing":
						case "line-height":
						case "margin-bottom":
						case "margin-left":
						case "margin-right":
						case "margin-top":
						case "max-height":
						case "max-width":
						case "min-height":
						case "min-width":
						case "padding-bottom":
						case "padding-left":
						case "padding-right":
						case "padding-top":
						case "right":
						case "top":
						case "width":
							o.style[coati_string.toCamelCase(animType)] = v + "px"; break;
					}
				});
			}

			// Runs the next animation step. (Internal use)
			//
			function doAnimate()
			{
				timePassed += 1000 / fps;
				if (timePassed >= duration)
				{
					timeoutHandle = undefined;
					self.finish(false);
					if (isFunction(self.onfinished))
						coati.defer(function() { self.onfinished(self); }, 1);   // Run asynchronously
					if (isObject(self.next) && "play" in self.next)
						coati.defer(self.next.play, 1);   // Run asynchronously
					return;
				}

				progress = (duration <= 0) ? 0 : timePassed / duration;
				doSet();

				timeoutHandle = setTimeout(doAnimate, 1000 / fps);

				if (isFunction(self.onstep))
					self.onstep(self);
			};

			var self = {
				// Public variables (properties)
				onfinished: undefined,
				onstep: undefined,
				next: undefined,

				// Gets the current progress of the animation (0..1).
				//
				progress: function()
				{
					return progress;
				},

				// Gets a value indicating whether the animation is currently playing.
				//
				playing: function()
				{
					return playing;
				},

				// Sets all specifications for this animation from the one it's attached to.
				//
				// This will only set the values that have not been set explicitly for this instance before.
				//
				setDefaults: function(d)
				{
					if ("o" in d && !coati_array_contains(ownSpecs, "o"))
					{
						o = coati_dom_selectAll(d.o);
						if (o.length == 1)
							o = o[0];
					}
					if ("type" in d && !coati_array_contains(ownSpecs, "type"))
						animType = d.type;
					if ("rel" in d && !coati_array_contains(ownSpecs, "rel"))
						endRelative = d.rel;
					if ("end" in d && !coati_array_contains(ownSpecs, "end"))
					{
						endVal = d.end;
						// startVal might have been auto-detected by the local endVal, but if it was originally
						// unset, the new endVal must trigger a new calculation of startVal now
						// (unless startVal is also set by the new specs, which is handled right below).
						if (!coati_array_contains(ownSpecs, "start"))
						{
							alert(1); // TODO: What's this?
							startVal = undefined;
						}
					}
					if ("start" in d && !coati_array_contains(ownSpecs, "start"))
						startVal = d.start;
					if ("duration" in d && !coati_array_contains(ownSpecs, "duration"))
						duration = d.duration;
					if ("method" in d && !coati_array_contains(ownSpecs, "method"))
						method = d.method;
					if ("param" in d && !coati_array_contains(ownSpecs, "param"))
						methodParam = d.param;
					if ("fps" in d && !coati_array_contains(ownSpecs, "fps"))
						self.fps = d.fps;
					findStartEnd();

					coati_array_each(attachedAnims, function(a) { a.setDefaults(d); });
				},

				// Attaches another Animation to this instance.
				//
				// The attached Animation's settings are filled in with this one's where they're yet unset.
				// Also, all playback commands and most configuration changes are passed on to all attached
				// Animations as well so that the group of animations appears to be a single complex animation.
				//
				attach: function(a)
				{
					var defaults = {
						o: o,
						type: animType,
						duration: duration,
						method: method,
						param: methodParam,
						fps: self.fps
					};
					if (coati_array_contains(ownSpecs, "start"))
						defaults.start = startVal;
					if (coati_array_contains(ownSpecs, "end"))
						defaults.end = endVal;
					if (coati_array_contains(ownSpecs, "rel"))
						defaults.rel = endRelative;

					if (!isArray(a)) a = [a];
					coati_array_each(a, function(a) {
						a.setDefaults(defaults);
						attachedAnims.push(a);
					});
					return self;
				},

				// Sets the total animation time in milliseconds.
				//
				setDuration: function(d)
				{
					duration = d;
					ownSpecs.push("duration");

					coati_array_each(attachedAnims, function(a) { a.setDuration(d); });
				},

				// Sets the animation movement type and its configuration parameter.
				//
				// m = (string) Movement type (see list above)
				// p = (number) Curve parameter [optional if unused by specified type]
				//
				setMethod: function(m, p)
				{
					method = m;
					methodParam = p;
					ownSpecs.push("method");
					ownSpecs.push("param");

					coati_array_each(attachedAnims, function(a) { a.setMethod(m, p); });
				},

				// Sets the animation smoothness as frames per second.
				//
				setFPS: function(f)
				{
					fps = f;
					ownSpecs.push("fps");

					coati_array_each(attachedAnims, function(a) { a.setFPS(f); });
				},

				// Starts the animation.
				//
				play: function()
				{
					self.reset();
					self.resume();
				},

				// Pauses the animation.
				//
				halt: function()
				{
					clearTimeout(timeoutHandle);
					playing = false;

					coati_array_each(attachedAnims, function(a) { a.halt(); });
				},

				// Resumes the animation.
				//
				resume: function()
				{
					if (duration <= 0) return false;
					if (fps < 1) return false;

					playing = true;
					doAnimate();

					coati_array_each(attachedAnims, function(a) { a.play(); });
				},

				// Resets the animation to its initial state.
				//
				reset: function()
				{
					progress = timePassed = 0;
					doSet();

					coati_array_each(attachedAnims, function(a) { a.reset(); });
				},

				// Sets the animation to its final state.
				//
				finish: function(cascade)
				{
					timePassed = duration;
					progress = 1;
					doSet();
					playing = false;

					if (!isSet(cascade) || cascade !== false)
						coati_array_each(attachedAnims, function(a) { a.finish(); });
				},

				// Reverses the animation while it's playing (or paused).
				//
				reverse: function()
				{
					var x = startVal;
					startVal = endVal;
					endVal = x;
					timePassed = duration - timePassed;
					progress = 1 - progress;

					coati_array_each(attachedAnims, function(a) { a.reverse(); });
				},

				// Gets the object(s) being animated.
				//
				getObject: function()
				{
					return o;
				}
			}; // coati.animation.self = {...}

			// Initialisation code

			// If an array was passed to the constructor, the first item of it is used for this instance
			// and all additional items are attached as separate Animations at the end of the constructor.
			if (isArray(params))
			{
				var firstParam = params.shift();
				additionalSpecs = params;
				params = firstParam;
			}

			// Get the values out of the passed object
			if ("o" in params)
			{
				ownSpecs.push("o");
				o = coati_dom_selectAll(params.o);
				if (o.length == 1)
					o = o[0];
			}
			if ("type" in params)
			{
				ownSpecs.push("type");
				animType = params.type;
			}
			if ("start" in params)
			{
				ownSpecs.push("start");
				startVal = params.start;
			}
			if ("end" in params)
			{
				ownSpecs.push("end");
				endVal = params.end;
			}
			if ("rel" in params)
			{
				ownSpecs.push("rel");
				endRelative = params.rel;
			}
			if ("duration" in params)
			{
				ownSpecs.push("duration");
				duration = params.duration;
			}
			if ("method" in params)
			{
				ownSpecs.push("method");
				method = params.method;
			}
			if ("param" in params)
			{
				ownSpecs.push("param");
				methodParam = params.param;
			}
			if ("fps" in params)
			{
				ownSpecs.push("fps");
				fps = params.fps;
			}
			if ("play" in params)
				autoPlay = params.play;
			if ("onfinished" in params)
				self.onfinished = params.onfinished;
			if ("onstep" in params)
				self.onstep = params.onstep;
			if ("next" in params)
				self.next = params.next;

			findStartEnd();

			// If multiple specifications were passed to the constructor, attach each of them
			// as a separate Animation to this instance.
			coati_array_each(additionalSpecs, function(params) {
				self.attach(new coati.animation(params));
			});

			if (autoPlay)
				self.play();

			return self;
		},


		// **********************************************************************
		// ***************                                        ***************
		// ***************             FORM HANDLING              ***************
		// ***************                                        ***************
		// **********************************************************************

		form: function(theForm)
		{
			theForm = coati_dom_select(theForm);

			function getElementValue(element)
			{
				switch (element.nodeName.toLowerCase())
				{
					case "input":
						switch (element.type)
						{
							case "button":
							case "hidden":
							case "password":
							case "submit":
							case "text":
							case "textarea":
								return element.value;
							case "checkbox":
								if (element.checked) return element.value;
								return "";
							case "radio":
								if (element.checked) return element.value;
								return undefined;   // try next element to find selected one
						}
						return undefined;
					case "textarea":
						return element.value;
					case "select":
						if (element.multiple)
						{
							var values = [];
							for (var i = 0; i < element.options.length; i++)
								if (element.options[i].selected)
									values.push(element.options[i].value);
							return values;
						}
						return element.value;
				}
				return undefined;
			}

			var self = {
				// Gets a field element node by its name.
				//
				// If the name is "submit", the form's submit button is returned.
				//
				getFieldElement: function(name)
				{
					for (var i = 0; i < theForm.elements.length; i++)
					{
						var el = theForm.elements[i];
						if (el.name == name)
							return el;
					}
					if (name == "submit")
					{
						for (var i = 0; i < theForm.elements.length; i++)
						{
							var el = theForm.elements[i];
							if (el.tagName.toLowerCase() == "input" && el.type.toLowerCase() == "submit")   // TODO: Could possibly also check for "image"
								return el;
						}
					}
					return undefined;
				},

				// Gets a field's value by its name.
				//
				getValue: function(name)
				{
					for (var i = 0; i < theForm.elements.length; i++)
					{
						var el = theForm.elements[i];
						if (el.name == name)
						{
							var v = getElementValue(el);
							if (isSet(v)) return v;
						}
					}
					return undefined;
				},

				// Gets all fields' values in an associative array.
				//
				getData: function(excludeNames)
				{
					var arr = {};
					for (var i = 0; i < theForm.elements.length; i++)
					{
						var el = theForm.elements[i];
						if (el.name != "" && !coati_array_contains(excludeNames, el.name))
						{
							if (!isSet(arr[el.name]))
							{
								arr[el.name] = getElementValue(el);
							}
							else
							{
								// This element name occurs more than once, make an array of it
								if (!isArray(arr[el.name]))
									arr[el.name] = [arr[el.name], getElementValue(el)];
								else
									arr[el.name].push(getElementValue(el));
							}
						}
					}
					return arr;
				},

				// Gets all fields' values in a single URL-encoded string.
				//
				getDataString: function(excludeNames)
				{
					var data = self.getData(excludeNames);
					var str = "";
					for (var prop in data)
					{
						var value = data[prop];
						if (isBoolean(value) || isNumber(value) || isString(value))
							str += (str != "" ? "&" : "") + prop + "=" + encodeURIComponent(value);
						else if (isArray(value))
							coati_array_each(value, function (v) {
								str += (str != "" ? "&" : "") + prop + "=" + encodeURIComponent(v);
							});
					}
					return str;
				},

				// Sets a field's new value by its name.
				//
				setValue: function(name, value)
				{
					for (var i = 0; i < theForm.elements.length; i++)
					{
						var el = theForm.elements[i];
						if (el.name == name)
						{
							switch (el.nodeName.toLowerCase())
							{
								case "input":
									switch (el.type)
									{
										case "button":
										case "hidden":
										case "password":
										case "submit":
										case "text":
										case "textarea":
											el.value = value;
											break;
										case "checkbox":
											el.checked = coati.isTrue(value);
											break;
										case "radio":
											el.checked = (el.value == value);   // check the one with the correct value or none
											break;
									}
									break;
								case "textarea":
									el.value = value;
									break;
								case "select":
									if (el.multiple)
									{
										for (var o = 0; o < el.options.length; o++)
										{
											el.options[o].selected = coati_array_contains(value, el.options[o].value);
										}
									}
									else
									{
										el.value = value;
									}
									break;
							}
							return;
						}
					}
					if (name == "submit")
					{
						for (var i = 0; i < theForm.elements.length; i++)
						{
							var el = theForm.elements[i];
							if (el.tagName.toLowerCase() == "input" && el.type.toLowerCase() == "submit")
							{
								el.value = value;
								return;
							}
						}
					}
				},

				// Removes all options from a selection list element.
				//
				removeAllOptions: function (name)
				{
					var select = self.getFieldElement(name);
					if (select.nodeName.toLowerCase() == "select")
						while (select.hasChildNodes())
							select.removeChild(select.firstChild);
				},

				// Adds a new option to a selection list element.
				//
				addOption: function (name, value, text)
				{
					var select = self.getField(name);
					if (select.nodeName.toLowerCase() == "select")
					{
						var option = document.createElement("option");
						var valueAttr = document.createAttribute("value");
						valueAttr.nodeValue = value;
						option.setAttributeNode(valueAttr);
						option.appendChild(document.createTextNode(text));
						select.appendChild(option);
					}
				},

				// Attaches a submit handler function to the form.
				//
				// The handler function will be passed the event and this Form object as
				// parameters.
				// The handler function can return a boolean value of false to indicate that
				// the form should not be submitted. If nothing or the value true is returned,
				// The form can be submitted.
				//
				// ckfn = (function) Handler function
				//
				addSubmitHandler: function(ckfn)
				{
					// Build a handler function
					var handler = function(event)
					{
						var res = ckfn(event, theForm);
						if (!isSet(res))
						{
							return true;
						}
						return res;
					};
					coati.dom.addEvent(theForm, "submit", handler);
				},

				// Attach a handler function to a form field
				//
				// The handler function is called after the field value has changed and thus
				// cannot be used to deny any field change.
				// The handler function will be passed the event and the field object as
				// parameters.
				// The handler function can return a non-empty string value that should be used as
				// the error message.
				//
				// name = (string) Field name
				// ckfn = (function) Handler function
				// onSubmit = (bool) true to check this field on submit time, false to check it on change
				// immediately = (bool) Check the field value already at call time (optional)
				//
				addFieldHandler: function(name, ckfn, onSubmit, immediately)
				{
					if (!isSet(immediately)) immediately = true;

					var popup = new coati.popup("popup-error");

					// Find the element and attach the event
					for (var i = 0; i < theForm.elements.length; i++)
					{
						if (theForm.elements[i].name == name)
						{
							var el = theForm.elements[i];

							// Build a handler function that handles the popup
							var handler = function(event)
							{
								var res = ckfn(event, el, theForm);
								if (isString(res) && res != "")
								{
									coati_dom.addClass(el, "checkError");
									if (res != "") popup.popup(el, res, 3);
									el.focus();
									return false;
								}
								else
								{
									coati_dom.removeClass(el, "checkError");
									popup.hide();
									return true;
								}
							};

							if (onSubmit)
							{
								coati_dom.addEvent(theForm, "submit", handler);
							}
							else
							{
								coati_dom.addEvent(el, "keyup", handler);
								coati_dom.addEvent(el, "click", handler);
								coati_dom.addEvent(el, "change", handler);

								// Call handler immediately
								if (immediately)
									setTimeout(function() { handler(null, el); }, 200);

								// Find surrounding <label> element
								var label = coati_dom.findParentByName(el, "label");
								if (label)
									coati_dom.addEvent(label, "click", handler);
							}
						}
					}
				}

			};
			return self;
		},


		// **********************************************************************
		// ***************                                        ***************
		// ***************                 POPUPS                 ***************
		// ***************                                        ***************
		// **********************************************************************

		popup: function(className)
		{
			// Private variables
			var node;
			var x = 0;
			var y = 0;
			var isVisible = true;
			var initialised = false;
			var hideTimeoutHandle;

			// Private methods

			// from Facebook:
			// Outer div: padding-{left|right|top|bottom}: 10px; position: absolute;
			// Main div: box-shadow: 0 3px 8px rgba(0, 0, 0, 0.3);
			// Inner span: position: absolute; {right|left|top|bottom}: 2px; {top|...}: 15px; background: {image}; width/height...;

			var self = {
				// Moves the popup to a new position.
				//
				move: function(newX, newY)
				{
					x = newX;
					y = newY;
					node.style.left = x + "px";
					node.style.top = y + "px";
				},

				// Moves the popup below an element.
				//
				moveBelow: function(obj)
				{
					self.move(coati_dom.left(obj) + 8, coati_dom.bottom(obj) - 4);
				},

				// Shows the popup.
				//
				show: function()
				{
					if (hideTimeoutHandle)
						clearTimeout(hideTimeoutHandle);
					isVisible = true;
					node.style.display = "";
				},

				// Hides the popup.
				//
				hide: function()
				{
					if (hideTimeoutHandle)
					{
						clearTimeout(hideTimeoutHandle);
						hideTimeoutHandle = undefined;
					}
					isVisible = false;
					new coati.animation({
						o: node,
						type: "opacity",
						start: 1,
						end: 0,
						play: true,
						onfinished: self.hideFast
					});
				},

				// Hides the popup without fading animation.
				//
				hideFast: function()
				{
					if (hideTimeoutHandle)
					{
						clearTimeout(hideTimeoutHandle);
						hideTimeoutHandle = undefined;
					}
					isVisible = false;
					node.style.display = "none";
					coati_dom.setOpacity(node, 1);
				},

				// Hides the popup after sec seconds.
				//
				hideTimeout: function(sec)
				{
					if (hideTimeoutHandle)
					{
						clearTimeout(hideTimeoutHandle);
						hideTimeoutHandle = undefined;
					}
					hideTimeoutHandle = setTimeout(self.hide, sec * 1000);
				},

				// Shows the popup, moves it below an element and hides it after sec seconds.
				//
				popup: function(obj, text, sec)
				{
					self.hideFast();
					self.setText(text);
					self.moveBelow(obj);
					self.show();
					self.hideTimeout(sec);
				},

				// Is the popup visible?
				//
				visible: function()
				{
					return isVisible;
				},

				// Sets the popup text contents, can contain HTML.
				//
				setText: function(text)
				{
					node.innerHTML = text;
					if (text == "")
					{
						self.hide();
					}
				}
			};

			// Initialisation code
			node = document.createElement("div");
			node.style.position = "absolute";
			coati_dom.addClass(node, "coati-popup");
			coati_dom.addClass(node, className);
			self.hideFast();
			coati.dom.whenDomLoaded(function () { document.body.appendChild(node); });
			initialised = true;

			return self;
		}

	}; // self = coati's object

	// Define private local variables for some commonly used functions.
	// These are then local names which can safely be shortened by a JavaScript compressor.
	// This is likely to amplify the savings effect in the compressed version compared to this
	// source file. There is no performance difference between both solutions.
	//
	var isSet = self.isSet;
	var isNumber = self.isNumber;
	var isBoolean = self.isBoolean;
	var isString = self.isString;
	var isArray = self.isArray;
	var isObject = self.isObject;
	var isFunction = self.isFunction;
	var coati_array = self.array;
	var coati_array_contains = self.array.contains;
	var coati_array_each = self.array.each;
	var coati_string = self.string;
	var coati_dom = self.dom;
	var coati_dom_select = self.dom.select;
	var coati_dom_selectAll = self.dom.selectAll;

	return self;
}();


// **********************************************************************
// ***************                                        ***************
// ***************            INITIALISATION              ***************
// ***************                                        ***************
// **********************************************************************

// Shortcuts for the DOM selection functions.
// You can safely remove or rename these names in case they are conflicting with something else.
$ = coati.dom.select;
$$ = coati.dom.selectAll;

// Make the whenLoaded event work properly.
coati.dom.whenDomLoaded(function() { coati.dom.setDomLoaded(); });
coati.dom.whenPageLoaded(function() { coati.dom.setPageLoaded(); });
